"""
.. note::
	It is recommended to not use assets directly. All property access is 
	handeled through a front facing ThingType. Assets are :mod:`UniqueThings <blueprints.thing.unique>`.
"""



import os
import struct
import xml.etree.ElementTree as xml
import numpy as np
import blueprints as blue
from blueprints.thing.colored import Color
from collections import defaultdict
from itertools import count
from imageio import imread, imwrite



ASSET_DIR = 'assets'



class BaseAsset(blue.AssetType, blue.UniqueThing):
	"""
	Assets are used to link external resources to the model. Currently only :class:`MeshAssets <MeshAsset>`, 
	 :class:`HFieldAssets <HFieldAsset>`, :class:`TextureAssets <TextureAsset>`, 
	 :class:`MaterialAssets <MaterialAsset>` are supported, but ``SkinAssets`` and 
	``BoneAssets`` will be developed in a future version version.
	"""
	@property
	def name(self) -> str:
		"""
		The user specified name might potentially be altered to avoid a naming conflict by appending an 
		enumeration scheme.
		
		Returns
		-------
		str
		"""
		return self._name or f'asset_{self.ID}'


	@name.setter
	@blue.restrict
	def name(self, name: str) -> None:
		"""
		The user specified name might potentially be altered to avoid a naming conflict by appending an 
		enumeration scheme.
		
		Parameters
		----------
		name : str
			The user specified name.
		"""
		self._name = name



class MaterialAsset(blue.MaterialAssetType, BaseAsset, blue.thing.ColoredThing):
	"""
	See :mod:`Material <blueprints.material>`
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#asset-material>`__.
	"""
	@blue.restrict
	def __init__(self, 
		     texture:     blue.TextureType|None = None, 
		     texrepeat:   list[int|float]|np.ndarray = [1, 1], 
		     texuniform:  bool      = False, 
		     emission:    int|float = 0.0, 
		     specular:    int|float = 0.5, 
		     shininess:   int|float = 0.5, 
		     reflectance: int|float = 0.0, 
		     metallic:    int|float =-1.0, 
		     roughness:   int|float =-1.0, 
		     color:       str|object|None = None, 
		     name:        str|None        = None):
		self._built      = False
		# MUJOCO ATTRIBUTES
		self.texrepeat   = texrepeat
		self.texuniform  = texuniform
		self.emission    = emission
		self.specular    = specular
		self.shininess   = shininess
		self.reflectance = reflectance
		self.metallic    = metallic
		self.roughness   = roughness
		# TEXTURE
		self.texture     = texture
		super().__init__(name=name)
		if color is not None:
			self.color = color
		if name is None:
			self._name = None


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		if not self._built:
			self._built = True
			if self.texture is not None:
				self._xml_root = xml.SubElement(world._xml_asset, 
							        'material', 
							        texture=self.texture.asset.name, 
								**self._mujoco_specs(kwargs))
			else:
				self._xml_root = xml.SubElement(world._xml_asset, 
							        'material', 
								**self._mujoco_specs(kwargs))
			#self._index       = indicies['mesh']
			#indicies['mesh'] += 1
			if self.texture is not None:
				self.texture._build(parent=parent, 
						    world=world, 
						    indicies=indicies, 
						    **kwargs)
		return self._xml_root



	@property
	def texrepeat(self) -> np.ndarray:
		"""
		The number of repetitions of the Texture in two directions

		Returns
		-------
		np.ndarray
		"""
		return self._texrepeat.copy()


	@texrepeat.setter
	@blue.restrict
	def texrepeat(self, texrepeat: list[int|float]|np.ndarray) -> None:
		if isinstance(texrepeat, list):
			texrepeat = np.array(texrepeat, dtype=np.float32)
		self._texrepeat = texrepeat


	@property
	def texuniform(self) -> bool:
		"""
		For :class:`BoxTextures <blueprints.texture.Box>`, this attribute controls how 
		cube mapping is applied. The default value ``False`` means apply cube mapping 
		directly, using the actual size of the object. The value “true” maps the texture 
		to a unit object before scaling it to its actual size (geometric primitives are 
		created by the renderer as unit objects and then scaled). In some cases this 
		leads to more uniform texture appearance, but in general, which settings produces 
		better results depends on the texture and the object. For 2d textures, this attribute 
		interacts with texrepeat above. Let texrepeat be N. The default value “false” means 
		that the 2d texture is repeated N times over the (z-facing side of the) object. 
		The value “true” means that the 2D :class:`PlaneTexture <blueprints.texture.Plane>` 
		is repeated N times over one spatial unit, regardless of object size.

		Returns
		-------
		bool
		"""
		return self._texuniform


	@texuniform.setter
	@blue.restrict
	def texuniform(self, texuniform: bool) -> None:
		self._texuniform = texuniform


	@property
	def emission(self) -> float:
		"""
		If set positive, the Material emitts lights.
		
		Returns
		-------
		float
		"""
		return self._emission


	@emission.setter
	@blue.restrict
	def emission(self, emission: int|float) -> None:
		self._emission = float(emission)


	@property
	def specular(self) -> float:
		"""
		Specularity in OpenGL. This value should be in the range [0 1].

		Returns
		-------
		float
		"""
		return self._specular


	@specular.setter
	@blue.restrict
	def specular(self, specular: int|float) -> None:
		self._specular = float(specular)


	@property
	def shininess(self) -> float:
		"""
		Shininess in OpenGL.

		Returns
		-------
		float
		"""
		return self._shininess


	@shininess.setter
	@blue.restrict
	def shininess(self, shininess: int|float) -> None:
		self._shininess = float(shininess)


	@property
	def reflectance(self) -> float:
		"""
		This attribute should be in the range [0 1]. If the value 
		is greater than 0, and the material is applied to a plane 
		or a box geom, the renderer will simulate reflectance. The 
		larger the value, the stronger the reflectance. For boxes, 
		only the face in the direction of the local +Z axis is 
		reflective. Simulating reflectance properly requires 
		ray-tracing which cannot (yet) be done in real-time. We 
		are using the stencil buffer and suitable projections instead. 
		Only the first reflective geom in the model is rendered as 
		such. This adds one extra rendering pass through all geoms, 
		in addition to the extra rendering pass added by each 
		shadow-casting light.

		Returns
		-------
		float
		"""
		return self._reflectance


	@reflectance.setter
	@blue.restrict
	def reflectance(self, reflectance: int|float) -> None:
		self._reflectance = float(reflectance)


	@property
	def metallic(self) -> float:
		"""
		This attribute corresponds to uniform metallicity 
		coefficient applied to the entire material. This 
		attribute has no effect in MuJoCo’s native renderer, 
		but it can be useful when rendering scenes with a 
		physically-based renderer. In this case, if a 
		non-negative value is specified, this metallic value 
		should be multiplied by the metallic texture sampled 
		value to obtain the final metallicity of the material.

		Returns
		-------
		float
		"""
		return self._metallic


	@metallic.setter
	@blue.restrict
	def metallic(self, metallic: int|float) -> None:
		self._metallic = float(metallic)


	@property
	def roughness(self) -> float:
		"""
		This attribute corresponds to uniform roughness 
		coefficient applied to the entire material. This 
		attribute has no effect in MuJoCo’s native renderer, 
		but it can be useful when rendering scenes with a 
		physically-based renderer. In this case, if a non-negative 
		value is specified, this roughness value should be 
		multiplied by the roughness texture sampled value to 
		obtain the final roughness of the material.

		Returns
		-------
		float
		"""
		return self._roughness


	@roughness.setter
	@blue.restrict
	def roughness(self, roughness: int|float) -> None:
		self._roughness = float(roughness)



class TextureAsset(blue.TextureAssetType, BaseAsset):
	"""
	See :mod:`Texture <blueprints.texture>`
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#asset-texture>`__.
	"""
	@blue.restrict
	def __init__(self, 
		     #color_space: str       = 'auto', 
		     content:     str|None  =  None, 
		     file:        str|None  =  None, 
		     grid_size:   list[int] = [1, 1], 
		     grid_layout: str|None  =  None,
		     fileright:   str|None  =  None, 
		     fileleft:    str|None  =  None, 
		     fileup:      str|None  =  None, 
		     filedown:    str|None  =  None, 
		     filefront:   str|None  =  None, 
		     fileback:    str|None  =  None, 
		     builtin:     str|None  =  None, 
		     color_1:     object    = [0.8, 0.8, 0.8], 
		     color_2:     object    = [0.5, 0.5, 0.5], 
		     mark:        str|None  =  None, 
		     color_mark:  object    = [0.0, 0.0, 0.0],
		     random:      float     =  0.01, 
		     width:       int       =  0, 
		     height:      int       =  0, 
		     h_flip:      bool      =  False, 
		     v_flip:      bool      =  False, 
		     n_channel:   int       =  3, 
		     name:        str|None  = None):
		self._built       = False
		self._image       = None
		self._image_up    = None
		self._image_down  = None
		self._image_left  = None
		self._image_right = None
		self._image_front = None
		self._image_back  = None
		#self.color_space = color_space
		self.content      = content
		self.file         = file
		self.grid_size    = grid_size
		self.grid_layout  = grid_layout
		self.fileright    = fileright
		self.fileleft     = fileleft
		self.fileup       = fileup
		self.filedown     = filedown
		self.filefront    = filefront
		self.fileback     = fileback
		self.builtin      = builtin
		self.color_1      = color_1
		self.color_2      = color_2
		self.mark         = mark
		self.color_mark   = color_mark
		self.random       = random
		self.width        = width
		self.height       = height
		self.h_flip       = h_flip
		self.v_flip       = v_flip
		self.n_channel    = n_channel
		super().__init__(name=name)
		self._load_images()
		if name is None:
			self._name = None


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		if not self._built:
			# MAKE DIRECTORIES
			assets_path = f'{world._path}/{ASSET_DIR}'
			if not os.path.exists(assets_path):
				os.mkdir(assets_path)
			dirname = f'{world._path}/{ASSET_DIR}/textures'
			if not os.path.exists(dirname):
				os.mkdir(dirname)
			# SAVE FILES
			filenames = self._save_images(dirname)
			# UPDATE FILENAMES
			kwargs = kwargs.copy()
			kwargs.update(filenames)
			self._built = True
			self._xml_root = xml.SubElement(world._xml_asset, 
							'texture', 
						        **self._mujoco_specs(kwargs))
		return self._xml_root


	def _load_images(self):
		if self.file       is not None:
			self._image       = imread(self.file)
		if self.fileup    is not None:
			self._imageup    = imread(self.fileup)
		if self.filedown  is not None:
			self._imagedown  = imread(self.filedown)
		if self.fileleft  is not None:
			self._imageleft  = imread(self.fileleft)
		if self.fileright is not None:
			self._imageright = imread(self.fileright)
		if self.filefront is not None:
			self._imagefront = imread(self.filefront)
		if self.fileback  is not None:
			self._imageback  = imread(self.fileback)


	def _save_images(self, dirname) -> dict:
		files = dict()
		if self.file       is not None:
			file       = self._resolve_filename(dirname, self.file)
			imwrite(file, self._image)
			files['file']       = file
		if self.fileup    is not None:
			fileup    = self._resolve_filename(dirname, self.fileup)
			imwrite(fileup, self._imageup)
			files['fileup']    = fileup
		if self.filedown  is not None:
			filedown  = self._resolve_filename(dirname, self.filedown)
			imwrite(filedown, self._imagedown)
			files['filedown']  = filedown
		if self.fileleft  is not None:
			fileleft  = self._resolve_filename(dirname, self.fileleft)
			imwrite(fileleft, self._imageleft)
			files['fileleft']  = fileleft
		if self.fileright is not None:
			fileright = self._resolve_filename(dirname, self.fileright)
			imwrite(fileright, self._imageright)
			files['fileright'] = fileright
		if self.filefront is not None:
			filefront = self._resolve_filename(dirname, self.filefront)
			imwrite(filefront, self._imagefront)
			files['filefront'] = filefront
		if self.fileback  is not None:
			fileback  = self._resolve_filename(dirname, self.fileback)
			imwrite(fileback, self._imageback)
			files['fileback']  = fileback
		return files


	def _resolve_filename(self, dirname: str, filename: str, prefix=None):
		pathname, basename = os.path.split(filename)
		if not pathname.endswith(dirname):
			if prefix is None:
				return f'{dirname}/{self.ID}_{basename}'
			else:
				return f'{dirname}/{self.ID}_{prefix}_{basename}'
		else:
			return filename

	# DERIVED ATTRIBUTES

	@property
	def type(self) -> str:
		"""
		Derived type attribute for Mujoco,

		Returns
		-------
		str
		"""
		return self._TYPE


	#@property
	#def colorspace(self) -> str:
	#	return self.color_space


	@property
	def content_type(self) -> str|None:
		"""
		Content type (MIME type) taking values ``None``, ``'image/png'``, 
		``'image/ktx'`` and ``'image/vnd.mujoco.texture'``.

		Returns
		-------
		str | None
		"""
		return self.content


	@property
	def gridsize(self) -> list[int]:
		"""
		Grid size attrigute ``[n_rows, n_cols]`` defining the grid. 
		For non-grid Textures this value is ignored.

		Returns
		-------
		list[int]
		"""
		return np.array(self.grid_size, dtype=np.int8)


	@property
	def gridlayout(self) -> str|None:
		"""
		The layout of the grid. See example above.

		Returns
		-------
		str | None
		"""
		return self.grid_layout


	@property
	def rgb1(self) -> np.ndarray:
		"""
		The first color of a procedural Texture.

		Returns
		-------
		blue.ColorType
		"""
		return self.color_1.rgb


	@property
	def rgb2(self) -> np.ndarray:
		"""
		The second color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self.color_2.rgb


	@property
	def markrgb(self) -> np.ndarray:
		"""
		The mark color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self.color_mark.rgb


	@property
	def hflip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the horizontal 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self.h_flip


	@property
	def vflip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the vertical 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self.v_flip


	@property
	def nchannel(self) -> int:
		"""
		The number of channels in the texture image file. This allows 
		loading 4-channel textures (RGBA) or single-channel textures 
		(e.g., for Physics-Based Rendering properties such as roughness 
		or metallic).

		Returns
		-------
		int
		"""
		return self.n_channel

	# BLUEPRINTS ATTRIBUTES

	#@property
	#def color_space(self) -> str:
	#	return self._color_space


	#@color_space.setter
	#@blue.restrict
	#def color_space(self, color_space: str) -> None:
	#	color_space = color_space.strip()
	#	if color_space not in ('auto', 'linear', 'sRGB'):
	#		raise ValueError(f'Attribute color_space must be set to "auto", "linear" or "sRGB". Got "{color_space}" instead.')
	#	self._color_space = color_space


	@property
	def content(self) -> str|None:
		"""
		Content type (MIME type) taking values ``None``, ``'image/png'``, 
		``'image/ktx'`` and ``'image/vnd.mujoco.texture'``.

		Returns
		-------
		str | None
		"""
		return self._content


	@content.setter
	@blue.restrict
	def content(self, content: str|None) -> None:
		if isinstance(content, str):
			content = content.strip()
		if content is not None and content not in ('image/png', 'image/ktx', 'image/vnd.mujoco.texture'):
			raise ValueError(f'Attribute content must be set to MIME type "image/png", "image/ktx" or "image/vnd.mujoco.texture". Got "{content}" instead.')
		self._content = content


	@property
	def file(self) -> str|None:
		"""
		Name of the texture file.

		Returns
		-------
		str | None
		"""
		return self._file


	@file.setter
	@blue.restrict
	def file(self, file: str|None) -> None:
		if isinstance(file, str):
			file = file.strip()
		self._file = file


	@property
	def grid_size(self) -> list[int]:
		"""
		Grid size attrigute ``[n_rows, n_cols]`` defining the grid. 
		For non-grid Textures this value is ignored.

		Returns
		-------
		list[int]
		"""
		return [self._n_rows, self._n_cols]


	@grid_size.setter
	@blue.restrict
	def grid_size(self, grid_size: list[int]) -> None:
		if len(grid_size) != 2:
			raise ValueError(f'grid_size must be a list of length 2. Got {len(grid_size)} instead.')
		rows, cols = grid_size
		if rows < 0:
			raise ValueError(f'Rows in grid_size[0] must be positive. Got {rows} instead.')
		if cols < 0:
			raise ValueError(f'Columns in grid_size[1] must be positive. Got {cols} instead.')
		if rows * cols > 12:
			raise ValueError(f'Rows × Cols in grid_size must not multiply beyond 12. Got {rows} × {cols} = {rows * cols} instead.')
		self._n_rows = rows
		self._n_cols = cols


	@property
	def grid_layout(self) -> str|None:
		"""
		The layout of the grid. See example above.

		Returns
		-------
		str | None
		"""
		return self._grid_layout


	@grid_layout.setter
	@blue.restrict
	def grid_layout(self, grid_layout: str|None) -> None:
		self._grid_layout = grid_layout


	@property
	def fileright(self) -> str|None:
		"""
		The file name for the right side of a :class:`Box`.

		Returns
		-------
		str | None
		"""
		return self._fileright


	@fileright.setter
	@blue.restrict
	def fileright(self, fileright: str|None) -> None:
		if isinstance(fileright, str):
			fileright = fileright.strip()
		self._fileright = fileright


	@property
	def fileleft(self) -> str|None:
		"""
		The file name for the left side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self._fileleft


	@fileleft.setter
	@blue.restrict
	def fileleft(self, fileleft: str|None) -> None:
		if isinstance(fileleft, str):
			fileleft = fileleft.strip()
		self._fileleft = fileleft


	@property
	def fileup(self) -> str|None:
		"""
		The file name for the upper side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self._fileup


	@fileup.setter
	@blue.restrict
	def fileup(self, fileup: str|None) -> None:
		if isinstance(fileup, str):
			fileup = fileup.strip()
		self._fileup = fileup


	@property
	def filedown(self) -> str|None:
		"""
		The file name for the down side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self._filedown


	@filedown.setter
	@blue.restrict
	def filedown(self, filedown: str|None) -> None:
		if isinstance(filedown, str):
			filedown = filedown.strip()
		self._filedown = filedown


	@property
	def filefront(self) -> str|None:
		"""
		The file name for the front side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self._filefront


	@filefront.setter
	@blue.restrict
	def filefront(self, filefront: str|None) -> None:
		if isinstance(filefront, str):
			filefront = filefront.strip()
		self._filefront = filefront


	@property
	def fileback(self) -> str|None:
		"""
		The file name for the back side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self._fileback


	@fileback.setter
	@blue.restrict
	def fileback(self, fileback: str|None) -> None:
		if isinstance(fileback, str):
			fileback = fileback.strip()
		self._fileback = fileback


	@property
	def builtin(self) -> str|None:
		"""
		If set, this attribute specifies a builtin procedural Texture. 
		See examples above.

		Returns
		-------
		str | None
		"""
		return self._builtin


	@builtin.setter
	@blue.restrict
	def builtin(self, builtin: str|None) -> None:
		if isinstance(builtin, str):
			builtin = builtin.strip()
		if builtin is not None and builtin not in ('gradient', 'checker', 'flat'):
			raise ValueError(f'Attribute builtin must be set to "gradient", "checker", "flat" or None. Got "{builtin}" instead.')
		self._builtin = builtin


	@property
	def color_1(self) -> Color:
		"""
		The first color of a procedural Texture.

		Returns
		-------
		blue.ColorType
		"""
		return self._color_1


	@color_1.setter
	@blue.restrict
	def color_1(self, color_1: object) -> None:
		self._color_1 = Color(color_1)


	@property
	def color_2(self) -> Color:
		"""
		The second color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self._color_2


	@color_2.setter
	@blue.restrict
	def color_2(self, color_2: object) -> None:
		self._color_2 = Color(color_2)


	@property
	def mark(self) -> str|None:
		"""
		If set this attribute specifies additional markings on 
		a procedural Texture. Possible values are ``None``, 
		``'edge'``, ``'cross'`` and ``'random'``. See examples above.
		
		Returns
		-------
		str | None
		"""
		return self._mark


	@mark.setter
	@blue.restrict
	def mark(self, mark: str|None) -> None:
		if isinstance(mark, str):
			mark = mark.strip()
		if mark is not None and mark not in ('edge', 'cross', 'random'):
			raise ValueError(f'Attribute mark must be set to "edge", "cross", "random" or None. Got "{mark}" instead.')
		self._mark = mark


	@property
	def color_mark(self) -> Color:
		"""
		The mark color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self._color_mark


	@color_mark.setter
	@blue.restrict
	def color_mark(self, color_mark: object) -> None:
		self._color_mark = Color(color_mark)


	@property
	def random(self) -> float:
		"""
		If the ``mark`` attribute is set to ``'random'``, this 
		attribute specifies the probability of a pixel in the 
		texture being the ``mark_color``.

		Returns
		-------
		float
		"""
		return self._random


	@random.setter
	@blue.restrict
	def random(self, random: float) -> None:
		if not 0 <= random <= 1:
			raise ValueError(f'Attribute random specifies a probability and must lay within the interval [0, 1]. Got {random} instead.')
		self._random = random


	@property
	def width(self) -> int:
		"""
		The width of a procedural texture, i.e., the number of columns 
		in the image. Larger values usually result in higher quality 
		images, although in some cases (e.g. checker patterns) small 
		values are sufficient. For textures loaded from files, this 
		attribute is ignored.

		Returns
		-------
		int
		"""
		return self._width


	@width.setter
	@blue.restrict
	def width(self, width: int) -> None:
		if width < 0:
			raise ValueError(f'Attribute width must be positive. Got {width} instead.')
		self._width = width


	@property
	def height(self) -> int:
		"""
		The height of the procedural texture, i.e., the number of rows 
		in the image. For :class:`Box` and :class:`Skybox` textures, this attribute is 
		ignored and the height is set to 6 times the width. For textures 
		loaded from files, this attribute is ignored.

		Returns
		-------
		int
		"""
		return self._height


	@height.setter
	@blue.restrict
	def height(self, height: int) -> None:
		if height < 0:
			raise ValueError(f'Attribute height must be positive. Got {height} instead.')
		self._height = height


	@property
	def n_channel(self) -> int:
		"""
		The number of channels in the texture image file. This allows 
		loading 4-channel textures (RGBA) or single-channel textures 
		(e.g., for Physics-Based Rendering properties such as roughness 
		or metallic).

		Returns
		-------
		int
		"""
		return self._n_channel


	@n_channel.setter
	@blue.restrict
	def n_channel(self, n_channel: int) -> None:
		if n_channel < 0:
			raise ValueError(f'Attribute n_channel must be positive. Got {n_channel} instead.')
		self._n_channel = n_channel


	@property
	def h_flip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the horizontal 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self._h_flip


	@h_flip.setter
	@blue.restrict
	def h_flip(self, h_flip: bool) -> None:
		self._h_flip = h_flip


	@property
	def v_flip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the vertical 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self._v_flip


	@v_flip.setter
	@blue.restrict
	def v_flip(self, v_flip: bool) -> None:
		self._v_flip = v_flip



class MeshAsset(blue.MeshAssetType, BaseAsset, blue.thing.MoveableThing):

	"""
	MeshAssets handle access to the data attributes used to define the :class:`Mesh <blueprints.geoms.Mesh>`. 
	All vertex, face, normals and texture coordinate data are stored in the MeshAssets :class:`MeshCache <blueprints.cache.MeshCache>` 
	in the :attr:`cache` attribute.

	In general, all Mesh data that are structured for xml representation and insufficient for user 
	manipulation are named as singular :attr:`vertex`, :attr:`face`, :attr:`texcoord` and :attr:`normal` 
	while their plural form :attr:`vertecies`, :attr:`faces`, :attr:`cache.texcoords <blueprints.cache.MeshCache.texcoords>` and :attr:`cache.normals <blueprints.cache.MeshCache.normals>` are used 
	for user interpretable representations and manipulation.

	.. note::
		The position of MeshAssets is defined in mujoco via the :attr:`refpos` attribute, which differs 
		to the standard ``pos`` attribute in that it defines the reference frames relative position w.r.t. 
		to its position instead of the vice versa the relative position w.r.t. the reference frames 
		origin, effectively making ``refpos = - pos``. We find this to be unnecessarily confusing and 
		opted to implement the standard position attribute only translating when the kinematic tree gets 
		converted to Mujoco xml. It is highly recommended to use the :attr:`pos` argument to modify the 
		MeshAssets position.
	
	Attributes
	----------
	built : bool
		A flag indicating whether the MeshAssets has been built. This is necessary since multiple 
		:class:`Meshes <blueprints.geoms.Mesh>` might use the same :class:`MeshAsset` in which case 
		it should be built only once.
	cache : blue.CacheType
		The :class:`MeshCache <blueprints.cache.MeshCache>` which stores the Meshs data.
	filename : str
		The filename from which the Mesh data is loaded or to which it will be saved.
	xml_data : bool
		A flag indicating whether the Mesh data is written into the xml directly or into and external 
		file which is then linked to in the xml asset element. 
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:       np.ndarray|list[int|float] = [0., 0., 0.],
		     vertecies: np.ndarray|list[np.ndarray|list[float|int]]|None = None,  
		     faces:     list[np.ndarray|list[float|int]]|None            = None,
		     scale:     np.ndarray|list[int|float] = [1., 1., 1.], 
		     filename:  str|None            = None, 
		     centered:  bool                = False, 
		     xml_data:  bool                = False, 
		     cache:     blue.ThingType|None = None, 
		     name:      str|None            = None, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			pos : np.ndarray | list[int | float], optional
			Represents the position of the Asset. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		vertecies : np.ndarray | list[np.ndarray | list[float | int]] | None, optional
			A list of all vertex positions. The vertecies are stored in :attr:`cache`.
		faces : list[np.ndarray | list[float | int]] | None, optional
			A list of all faces. A face consists of at least 3 indecies of vertecies. The faces are 
			stored in :attr:`cache`.
		scale : np.ndarray | list[int | float], optional
			The scale is a 1D array like object of length 3. Each component is used to scale the X, 
			Y and Z-axis respectively. Changing the scale is computationally cheap since it does not 
			change the vertecies or other data in :attr:`cache`.
		filename : str | None, optional
			The filename from which the mesh data is loaded and to to which the mesh will be saved. 
			The file is saved in a special directory such that no input file is overwritten by the 
			output file.
		centered : bool, optional
			If ``True`` the vertecies are normalized such that their mean position is the reference 
			frames origin.
		xml_data : bool, optional
			If ``True`` the Mesh data is not written to an external file and linked to in the xml, 
			but written into the xml directly. It is recommended to only use this option if the Mesh 
			data is small enough to not blot the xml.
		cache : blue.ThingType | None, optional
			The :class:`Cache <blueprints.cache.MeshCache>` which stores the Mesh data.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by 
			appending an enumeration scheme.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		if cache is None:
			self.cache = blue.cache.MeshCache(vertecies=vertecies, 
							  faces=faces, 
							  filename=filename, 
							  centered=centered)
		else:
			self.cache = cache
		self.cache._add(self)
		self.xml_data = xml_data
		#self.filename = filename
		self.scale    = scale
		self._built    = False
		super().__init__(pos=pos, 
				 name=name, 
				 **kwargs)
		#blue.REGISTER.assets.append(self)
		if name is None:
			self._name = None


	@blue.restrict
	def load(self, filename: str) -> None:
		"""
		This methods loads the mesh data from the filename to the assets cache.
		
		Parameters
		----------
		filename : str
			The filename must be binary or ascii for OBJ files or binary for STL files.
		"""
		self.cache.load(filename)


	@blue.restrict
	def save(self, filename: str) -> None:
		"""
		This method saves the mesh data from the cache to the filename.
		
		Parameters
		----------
		filename : str
			The filename must either end with the ``'.obj`` or ``'.stl'`` file extension. Both file 
			formats are saved as binary files.
		"""
		self.cache.save(filename)


	@blue.restrict
	@classmethod
	def _from_xml_element(cls, 
			      xml_element: xml.Element, 
			      cache:       blue.CacheType) -> blue.ThingType:
		"""
		This method reconstructs an MeshAsset from an xml element.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a MeshAsset is reconstructed.
		cache : blue.CacheType
			The Cache which stores the Mesh data.
		
		Returns
		-------
		blue.ThingType
			The reconstructed MeshAsset.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['cache'] = cache
		if 'refpos' in post_args:
			init_args['pos'] = - post_args.pop('refpos')
		if 'vertex' in rest_args:
			init_args['xml_data']  = True
			init_args['vertecies'] = rest_args.pop('vertex')
		else:
			init_args['xml_data']  = False
		mesh = object.__new__(cls)
		mesh.__init__(**init_args)
		for key, val in post_args.items():
			setattr(mesh, key, val)
		return mesh


	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called by the Assets parent to construct the xml representations of 
		the kinematic tree.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		**kwargs
			Keyword arguments are overwrite the mujoco specs for building.
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The built xml element of the Asset.
		"""
		if not self._built:
			if self.xml_data:
				mujoco_specs = self._mujoco_specs()
				mujoco_specs.update(self.cache._mujoco_specs())
				self._xml_root = xml.Element('mesh', **mujoco_specs)
				self.cache._ROOT = world
			else:
				# MAKE DIRECTORIES
				assets_path = f'{world._path}/{ASSET_DIR}'
				if not os.path.exists(assets_path):
					os.mkdir(assets_path)
				dirname = f'{world._path}/{ASSET_DIR}/meshes'
				if not os.path.exists(dirname):
					os.mkdir(dirname)
				# BUILD CACHE AND SELF
				self.cache._build(dirname)
				self._xml_root = xml.Element('mesh', 
							     file=self.cache._path, 
							     **self._mujoco_specs({'vertex': None, 'face': None}))
			self._index = indicies['mesh']
			self._built = True
		if self._xml_root not in world._xml_asset:
			world._xml_asset.append(self._xml_root)
		return self._xml_root

	# MUJOCO PROPERTIES

	@property
	def file(self) -> str | None:
		"""
		Derived dummy variable for Mujoco XML

		Returns
		-------
		str
		"""
		return self.filename


	@property
	def filename(self) -> str | None:
		"""
		The filename of the Mesh

		Returns
		-------
		str
		"""
		return self.cache.filename


	@filename.setter
	@blue.restrict
	def filename(self, filename: str | None) -> None:
		"""
		Returns
		-------
		str
		"""
		self.cache.filename = filename


	@property
	def refpos(self) -> np.ndarray:
		"""
		The negative :attr:`pos`. This attribute is implemented for mujoco, but synchronized with 
		:attr:`pos` which is recommended to to be used instead.
		
		Returns
		-------
		np.ndarray
		"""
		return - self.pos


	@property
	@blue.restrict
	def vertex(self) -> np.ndarray:
		"""
		This attribute is used to write construct the mujoco xml. It returns a flattened list of 
		coordinates that is written raw to xml. For a structured list of vertecies see :attr:`vertecies`.

		Returns
		-------
		np.ndarray
		"""
		return self.cache.vertex


	@property
	@blue.restrict
	def face(self) -> np.ndarray|None:
		"""
		This attribute is used to write to xml. It returns a flattened list of indecies that is 
		written raw to xml. For a structured list of faces see :attr:`faces`.

		Returns
		-------
		np.ndarray | None
		"""
		return self.cache.face


	@face.setter
	@blue.restrict
	def face(self, face: list) -> None:
		"""
		Parameters
		----------
		face : list
			This attribute is used to write to xml. It returns a flattened list of indecies that is 
			written raw to xml. For a structured list of faces see :attr:`faces`.
		"""
		self.cache._prepare_for_modification(self)
		self.cache.face = face


	@property
	def texcoord(self) -> np.ndarray|None:
		"""
		This attribute is used to write to xml. It returns a flattened list of texture coordinates that 
		is written raw to xml. For a structured list of texture coordinates see :attr:`texcoords`.
		
		Returns
		-------
		np.ndarray | None
		"""
		return self.cache.texcoord


	@texcoord.setter
	@blue.restrict
	def texcoord(self, texcoord: list) -> None:
		"""
		Parameters
		----------
		texcoord : list
			This attribute is used to write to xml. It returns a flattened list of texture 
			coordinates that is written raw to xml. For a structured list of texture coordinates see 
			:attr:`texcoords`.
		"""
		self.cache._prepare_for_modification(self)
		self.cache.texcoord = texcoord


	@property
	def normal(self) -> np.ndarray|None:
		"""
		This attribute is used to write to xml. It returns a flattened list of face normals that is 
		written raw to xml. For a structured list of face normals see :attr:`normals`.
		
		Returns
		-------
		np.ndarray | None
		"""
		return self.cache.normal


	@normal.setter
	@blue.restrict
	def normal(self, normal: list) -> None:
		"""
		Parameters
		----------
		normal : list
			This attribute is used to write to xml. It returns a flattened list of face normals that 
			is written raw to xml. For a structured list of face normals see :attr:`normals`.
		"""
		self.cache._prepare_for_modification(self)
		self.cache.normal = normal


	@property
	def size(self) -> np.ndarray:
		"""
		The size attribute specifies the ranges for each axis that the Mesh occupies.
		
		Returns
		-------
		np.ndarray
		"""
		return self.scale * self.cache.size


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]) -> None:
		"""
		The size attribute specifies the ranges for each axis that the Mesh occupies.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			Changing this parameter scales the vertecies directly in the :class:`Cache <blueprints.cache.MeshCache>` 
			which is computationally costly. To scale the size of the Mesh independently from the 
			vertecies in a computationally cheap manner use the :attr:`scale` attribute.
		"""
		self.cache._prepare_for_modification(self)
		size = np.array(size, dtype=np.float32) / self.scale
		size[np.isinf(size)] = 0
		self.cache.size = size


	@property
	def scale(self) -> np.ndarray:
		"""
		The scale is a 1D array like object of length 3. Each component is used to scale the X, Y and 
		Z-axis respectively. Changing the scale is computationally cheap since it does not change the 
		vertecies or other data in :attr:`cache`.
		
		Returns
		-------
		np.ndarray
		"""
		return self._scale


	@scale.setter
	@blue.restrict
	def scale(self, scale: np.ndarray|list[int|float]):
		"""
		Parameters
		----------
		scale : np.ndarray | list[int | float]
			The scale is a 1D array like object of length 3. Each component is used to scale the X, 
			Y and Z-axis respectively. Changing the scale is computationally cheap since it does not 
			change the vertecies or other data in :attr:`cache`.
		
		Raises
		------
		ValueError
			If the scale contains a component that is 0 an error is raised.
		"""
		scale = np.array(scale, dtype=np.float32)
		if np.any(scale == 0):
			raise ValueError(f'A meshes scale is not allowed to be zero in any component got {scale}.')
		self._scale = scale

	# BLUEPRINTS PROPERTIES

	@property
	def vertecies(self) -> list[np.ndarray]:
		"""
		The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
		spatial dimensions.
		
		Returns
		-------
		list[np.ndarray]
		"""
		return self.cache.vertecies


	@vertecies.setter
	@blue.restrict
	def vertecies(self, vertecies: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]]
			The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
			spatial dimensions.
		"""

		self.cache._prepare_for_modification(self)
		self.cache.vertecies = vertecies

	
	@property
	def faces(self) -> list[np.ndarray]|None:
		"""
		The list of faces. Each face is a np.ndarray with n indecies for each of the vertecies of the 
		face.
		
		Returns
		-------
		list[np.ndarray] | None
		"""
		return self.cache.faces


	@faces.setter
	@blue.restrict
	def faces(self, faces: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Parameters
		----------
		faces : np.ndarray | list[np.ndarray | list[float | int]]
			The list of faces. Each face is a np.ndarray with n indecies for each of the vertecies 
			of the face.
		"""
		self.cache._prepare_for_modification(self)
		self.cache.faces = faces



class HFieldAsset(blue.HFieldAssetType, BaseAsset, blue.thing.MoveableThing):
	"""
	See :class:`HField <blueprints.geoms.HField>`
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-geom>`__.
	"""
	@blue.restrict
	def __init__(self, 
		     filename: str|None                        = None, 
		     pos:      np.ndarray|list[int|float]      = [0., 0., 0.], 
		     terrain:  np.ndarray|list[int|float]|None = None, 
		     x_length: int|float|None                  = 1, 
		     y_length: int|float|None                  = 1, 
		     z_length: int|float|None                  = 1, 
		     height_offset: int|float|None             = 1, 
		     xml_data: bool                            = False, 
		     cache:    blue.ThingType|None             = None, 
		     name:     str|None                        = None, 
		     **kwargs) -> None:
		if cache is None:
			self.cache = blue.cache.HFieldCache(terrain=terrain, 
							    filename=filename)
		else:
			self.cache = cache
		self.cache._add(self)
		self.xml_data = xml_data
		self.filename = filename
		self.x_length = x_length
		self.y_length = y_length
		self.z_length = z_length
		self.height_offset = height_offset
		self._built    = False
		super().__init__(pos=pos, 
				 name=name, 
				 **kwargs)
		if name is None:
			self._name = None


	@blue.restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called by the Assets parent to construct the xml representations of 
		the kinematic tree.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		**kwargs
			Keyword arguments are overwrite the mujoco specs for building.
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The built xml element of the Asset.
		"""
		if not self._built:
			if self.xml_data and False:
				mujoco_specs = self._mujoco_specs()
				mujoco_specs.update(self.cache._mujoco_specs())
				self._xml_root = xml.Element('hfield', **mujoco_specs)
			else:
				# MAKE DIRECTORIES
				if not os.path.exists(ASSET_DIR):
					os.mkdir(ASSET_DIR)
				assets_path = f'{world._path}/{ASSET_DIR}/hfields'
				if not os.path.exists(assets_path):
					os.mkdir(assets_path)
				# BUILD CACHE AND SELF
				self.cache._build(assets_path)
				self._xml_root = xml.Element('hfield', 
							      file=self.cache._path, 
							      **self._mujoco_specs({'elevation': None}))
			self.cache._world   = world
			self.cache._index   = indicies['hfield']
			indicies['hfield'] += 1
			self._built = True
		if self._xml_root not in world._xml_asset:
			world._xml_asset.append(self._xml_root)
		self.cache._ROOT = world
		return self._xml_root


	@blue.restrict
	@classmethod
	def _from_xml_element(cls,
			      xml_element: xml.Element,
			      cache:       blue.CacheType) -> blue.ThingType:
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['cache'] = cache
		# size is a read-only derived property; convert to init args
		size = post_args.pop('size', None)
		if size is not None:
			init_args['x_length']      = float(size[0]) * 2
			init_args['y_length']      = float(size[1]) * 2
			init_args['z_length']      = float(size[2]) * 2
			init_args['height_offset'] = float(size[3]) * 2
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		return obj


	def __getitem__(self,
			key: tuple[slice|int]) -> np.ndarray|np.float32:
		"""
		Returns
		-------
		np.ndarray | np.float32
			The values from the selected index/slice of the height field.
		"""
		return self.cache[key]


	def __setitem__(self, 
			key:   tuple[slice|int], 
			value: int|float|list[int|float]|np.ndarray) -> None:
		"""
		Parameters
		----------
		key : tuple[slice | int]
			The indecies/slices of acces
		value : int|float|list[int|float]|np.ndarray
			The value to be assigned in the selected parts of the field.
		"""
		self.cache._prepare_for_modification(self)
		self.cache[key] = value

	# MUJOCO PROPERTIES

	@property
	def size(self) -> np.ndarray:
		"""
		Derived size attribute for Mujoco XML.

		Returns
		-------
		np.ndarray
		"""
		return np.array([self.x_length/2, self.y_length/2, self.z_length/2, self.height_offset/2])


	@property
	def name(self) -> str:
		"""
		The user specified name might potentially be altered to avoid a naming conflict by appending an 
		enumeration scheme.
		
		Returns
		-------
		str
		"""
		return self._name or f'asset_{self.ID}'


	@name.setter
	@blue.restrict
	def name(self, name: str) -> None:
		"""
		The user specified name might potentially be altered to avoid a naming conflict by appending an 
		enumeration scheme.
		
		Parameters
		----------
		name : str
			The user specified name.
		"""
		self._name = name

	# BLUEPRINTS PROPERTIES
		
	@property
	def x_length(self):
		"""
		Length along the ``X``-axis of the HField.

		Returns
		-------
		float
		"""
		return self._x_length


	@property
	def y_length(self):
		"""
		Length along the ``Y``-axis of the HField.

		Returns
		-------
		float
		"""
		return self._y_length


	@property
	def z_length(self):
		"""
		Length along the ``Z``-axis of the HField.

		Returns
		-------
		float
		"""
		return self._z_length


	@property
	def height_offset(self):
		"""
		Height offset along the ``Z``-axis of the HField.

		Returns
		-------
		float
		"""
		return self._height_offset


	@x_length.setter
	@blue.restrict
	def x_length(self, x_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self._x_length = float(x_length)


	@y_length.setter
	@blue.restrict
	def y_length(self, y_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self._y_length = float(y_length)


	@z_length.setter
	@blue.restrict
	def z_length(self, z_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self._z_length = float(z_length)


	@height_offset.setter
	@blue.restrict
	def height_offset(self, height_offset: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self._height_offset = float(height_offset)
	

	@property
	def terrain(self) -> list[np.ndarray]:
		"""
		The terrain data of the HField
		
		Returns
		-------
		list[np.ndarray]
		"""
		return self.cache.terrain


	@terrain.setter
	@blue.restrict
	def terrain(self, terrain: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]]
		"""
		self.cache._prepare_for_modification(self)
		self.cache.terrain = terrain



ASSET_THINGS  = {'mesh':   MeshAsset, 
		 'hfield': HFieldAsset}



import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml



class FocalThing(blue.FocalThingType, base.BaseThing):
	"""
	This class is similar in behavior to :class:`CyclicalThing <blueprints.thing.cyclical.CyclicalThing>` 
	with the exception that it does not have a concrete parent. 
	FocalThings are the target of multiple incomming attribute 
	assignments from higher up in the hierarchy hence merging 
	different path and introducing cycles. To deal with copies 
	of kinematic graphs containing FocalThings, migration methods 
	similar to those in :class:`CyclicalThing <blueprints.thing.cyclical.CyclicalThing>` are defined.
	
	.. note::
		This Type is used in subclassing checks to summarize multile concrete ThingTypes.
	"""
	
	@restrict
	def __init__(self, *args, **kwargs) -> None:
		super().__init__(*args, **kwargs)



"""
The :class:`World` class hosts all Things in a Simulation. It handles the access to Mujoco by implementing a 
method, that converts the Constructed World to an xml string and a method used to launch a Mujoco viewer, that 
is especially useful during model construction for debugging. You can add and remove Things from the World with 
the same methods as in :class:`NodeThing <blueprints.thing.node.NodeThing>` i.e. :meth:`attach <blueprints.thing.node.NodeThing.attach>` 
and :meth:`detach <blueprints.thing.node.NodeThing.detach>`.

.. code-block:: python
	:caption: World attachment/detachment

	>>> world = blue.World()
	>>> teapot = blue.geoms.Mesh(filename='utahteapot.stl', pos=[0, 0, 1])
	>>> red_teapot = teapot.copy(color='red')
	>>> green_teapot = teapot.copy(color='green')
	>>> world.attach(red_teapot, green_teapot, copy=False)
	>>> world.geoms.color
	[Red[#FF0000], Green[#00FF00]]
	>>> world.detach(red_teapot)
	>>> world.geoms.color
	[Green[#00FF00]]

The :meth:`view <World.view>` method can be used. The perspective of the default camera is set such that all 
Things attached to the :class:`World` are in view.

.. code-block:: python
	:caption: World view

	>>> world.view()

Things that are attached with the ``background`` flag argument set to ``True`` will not be used to center the 
perspective of the camera.

.. code-block:: python
	:caption: World background

	>>> plane = blue.geoms.Plane()
	>>> world.attach(plane, background=True)
	>>> world.view()

"""

import xml.etree.ElementTree as xml
import numpy as np
import os
import sys
import time
import tqdm
import imageio
import mujoco
import mujoco.viewer
from mujoco.glfw import glfw
import blueprints as blue

from collections import defaultdict
from itertools import chain



class World(blue.WorldType, blue.thing.NodeThing):

	r"""
	This class is available through the shortcut :class:`blueprints.World <World>`.

	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#mjcf-reference>`__.
	
	Attributes
	----------
	angle : str
		Either ``'radian'`` or ``'degree'`` (avoid if possible).
	autolimits : bool
		Indicating whether limits are used or not.
	contact : bool
		Indicating whether collisions and contacts are ignored in the simulations physics.
	gravity : bool
		Indicating whether gravity exists.
	integrator : str
		The name of the integrator use, see :attr:`integrator` for a detailed description.
	viscosity : float
		The viscosity of the Worlds Things.
	"""

	@blue.restrict
	def __init__(self, 
		     name:       str            = 'anonymous_model', 
		     autolimits: bool           = True, 
		     viscosity:  float|int|None = 0.,
		     timestep:   float|int|None = 0.002,
		     integrator: str            = 'implicit',
		     cone:       str            = 'pyramidal', 
		     gravity:    bool           = True, 
		     contact:    bool           = True, 
		     angle:      str            = 'radian', 
		     texture:    blue.SkyboxTextureType|None = None, 
		     **kwargs):
		r"""
		Parameters
		----------
		name : str, optional
			The user specified name for the Actuator. In the case of a naming conflict the name will 
			be altered by an enumeration scheme.
		autolimits : bool, optional
			This attribute affects the behavior of attributes such as ``'limited'`` (on <body-joint> 
			``'forcelimited'``, ``'ctrllimited'``, and ``'actlimited'`` (on <actuator>). If ``True``, 
			these attributes are unnecessary and their value will be inferred from the presence of their 
			corresponding ``'range'`` attribute. If ``False``, no such inference will happen: For a Joint 
			to be limited, both limited=``True`` and range=``min max`` must be specified. In this mode, 
			it is an error to specify a range without a limit. 
		timestep : float | int | None, optional
			Simulation time step in seconds. This is the single most important parameter affecting 
			the speed-accuracy trade-off which is inherent in every physics simulation. Smaller values 
			result in better accuracy and stability. To achieve real-time performance, the time step 
			must be larger than the CPU time per step (or 4 times larger when using the RK4 integrator). 
			The CPU time is measured with internal timers. It should be monitored when adjusting the 
			time step.
		viscosity : float | int | None, optional
			Viscosity of the medium. This parameter is used to simulate viscous forces, which scale 
			linearly with velocity. In SI units the viscosity of air is around 0.00002 while the 
			viscosity of water is around 0.0009 depending on temperature. Setting viscosity to 0 
			disables viscous forces. Note that the default Euler integrator handles damping in the 
			:class:`Joints <blueprints.joint.BaseJoint>` implicitly – which improves stability and 
			accuracy. It does not presently do this with body viscosity. Therefore, if the goal is 
			merely to create a damped simulation (as opposed to modelling the specific effects of 
			viscosity), we recommend using joint damping rather than body viscosity, or switching to 
			the implicit or implicitfast :attr:`integrator`.
		cone : str, optional
			The type of contact friction cone. Elliptic cones are a better model of the physical reality, but pyramidal cones sometimes make the solver faster and more robust.
		integrator : str, optional
			This attribute selects the numerical integrator to be used. Currently the available 
			integrators are the semi-implicit Euler method ``'Euler'``, the fixed-step 4-th order 
			Runge Kutta method ``'RK4'``, the Implicit-in-velocity Euler method ``'implicit'``, and  
			``'implicitfast'``, which drops the Coriolis and centrifugal terms.
		gravity : bool, optional
			A flag indicating whether gravity exists. Be default the `gravitational constant <https://en.wikipedia.org/wiki/Gravitational_constant>`__. 
			is set to :math:`6.674 \times 10^{-11} N \frac{m^2}{kg^2}`.
		contact : bool, optional
			If ``False``, collisions and contacts are ignored in the simulations physics.
		angle : str, optional
			The angle measure either ``'radian'`` or ``'degree'``. 
			.. warning::
			
			    At the moment 'degree' is not supported for any manipulations. It is recommended to 
			    only use 'degree' if no orientation manipulation is performed on the model! To convert 
			    from degrees to radians use the multiplier :attr:`DEGREES_TO_RADIANS <blueprints.utils.geometry.DEGREES_TO_RADIANS>`
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.angle       = angle
		self.gravity     = gravity
		self.timestep    = float(timestep)
		self.viscosity   = float(viscosity)
		self.integrator  = integrator
		self.cone        = cone
		self.contact     = contact
		self.autolimits  = autolimits
		self.texture     = texture
		self._agents     = []
		self._geoms      = []
		self._bodies     = []
		self._sites      = []
		self._lights     = []
		self._cameras    = []
		self._background = []
		self._CHILDREN   = {'lights':  {'type':     blue.LightType, 
						'children': self._lights}, 
				    'geoms':   {'type':     blue.GeomType, 
						'children': self._geoms}, 
				    'sites':   {'type':     blue.SiteType, 
						'children': self._sites}, 
				    'cameras': {'type':     blue.CameraType, 
						'children': self._cameras}, 
				    'agents':  {'type':     blue.AgentType, 
				    		'children': self._agents}, 
				    'bodies':  {'type':     blue.BodyType, 
						'children': self._bodies}}
		self._name_manager = blue.naming.NameManager(self)
		self._built        = False
		self._viewer       = None
		self._recordings   = dict()
		super().__init__(name=name, 
				 **kwargs)

	# KINEMATIC TREE PROPERTIES/METHODS

	@blue.restrict
	def attach(self, 
		   *items:      list[blue.ThingType|blue.LatticeType|blue.ViewType], 
		   copy:       bool = False, 
		   background: bool = False) -> None:
		"""
		This method attaches *items to the parent Thing. If copy is set true, a copy of the items is created. 
		If copy is set to false, the kinematic graph might no longer be a tree resulting in infinite loops 
		or naming conflicts if the same item is included in the tree twice. All items of a specific type 
		can be accessed via the types attribute.
		
		Parameters
		----------
		*items : list[blue.ThingType]
			All items that are to be attached.
		copy : bool, optional
			A flag indicating whether a copy of the items should be attached.
		background : bool, optional
			A flag indicating whether the attached Things belong to the Worlds background.

		Raises
		------
		TypeError
			If an item is not a valid child an error is raised.
		"""
		views   =                   list(chain( *filter(lambda x:     isinstance(x, blue.ViewType),                     items)))
		lattice =                   list(chain( *filter(lambda x:     isinstance(x, blue.LatticeType),                  items)))
		items   = views + lattice + list(        filter(lambda x: not isinstance(x, (blue.ViewType, blue.LatticeType)), items))
		if copy:
			items = [item.copy() for item in items]
		super().attach(*items, copy=False)
		if background:
			self._background.extend(items)


	@blue.restrict
	def detach(self, *items: list[blue.ThingType]) -> None:
		"""
		This method is used to detach Things from the kinematic tree. They are afterwards 
		no longer present in this Worlds children attributes and their ``parent`` attribute 
		will be set to ``None``. 
		
		Parameters
		----------
		*items
			Things that are no longer children of this World can be passed, as long as their 
			type is a valid type for children of :class:`World`.
		
		Raises
		------
		TypeError
			If the arguments to this function are not valid children types an error is raised.
		"""
		super().detach(*items)
		for item in items:
			if item in self._background:
				self._background.remove(item)

	# XML PROPERTIES/METHODS

	@blue.restrict
	def build(self) -> None:
		"""
		This method constructs the xml structure of the kinematic tree. It is called before the World 
		is converted to an xml string if any modification to it was done since the last build. It is 
		not necessary for a user to call this method since it is implicitly called every time the build 
		needs to be updated.
		"""
		if self._built:
			return
		# REGISTERING NAMES
		self._name_manager.register()
		# BUILD DIRECTORY
		pathname   = os.path.dirname(sys.argv[0])
		self._path = os.path.abspath(pathname)
		# BUILD SIZE AND CENTER
		min_pos, max_pos    = self._location_range
		self._center        = (min_pos + max_pos)/2
		self._size          = np.max(max_pos - min_pos)/2
		# BUILD TOP LEVEL XML
		self._xml_root      = xml.Element('mujoco', model=self.name)
		self._xml_compiler  = xml.SubElement(self._xml_root, 
							 'compiler', 
							 angle=self.angle, 
							 autolimits=str(self.autolimits).lower(), 
							 #inertiafromgeom='true', 
							 coordinate='local')
		option = xml.SubElement(self._xml_root, 
					'option', 
					timestep=str(self.timestep), 
					viscosity=str(self.viscosity), 
					integrator=self.integrator, 
					cone=self.cone)
		#statistic = xml.SubElement(self._xml_root, 
		#			'statistic', 
		#			 meansize=".05"
		#			)
		flag   = xml.SubElement(option, 
					'flag', 
					gravity='disable' if not self.gravity else 'enable', 
					contact='disable' if not self.contact else 'enable')
		self._xml_asset     = xml.SubElement(self._xml_root, 'asset')
		self._xml_sensor    = xml.SubElement(self._xml_root, 'sensor')
		self._xml_actuator  = xml.SubElement(self._xml_root, 'actuator')
		self._xml_tendon    = xml.SubElement(self._xml_root, 'tendon')
		
		if self.all.cameras:
			max_width, max_height = np.max(self.all.cameras.resolution, axis=0)
		else:
			max_width, max_height = 480, 480
		self._gl_context = mujoco.GLContext(max_height, max_width)
		
		self._xml_visual    = xml.SubElement(self._xml_root, 'visual')
		self._xml_global    = xml.SubElement(self._xml_visual, 'global', offwidth=str(max_width), offheight=str(max_height))
		# BUILD WORLDBODY XML
		indicies = {'body':       1, 
			    'geom':       0, 
			    'jnt':        0, 
			    'light':      0, 
			    'mesh':       0, 
			    'site':       0, 
			    'sensors':    0, 
			    'activation': 0, 
			    'force':      0, 
			    'hfield':     0}
		self._xml_worldbody = xml.SubElement(self._xml_root, 'worldbody')
		self._build_children(parent=self._xml_worldbody, 
				     world=self, 
				     indicies=indicies)
		if self.texture is not None:
			self.texture._build(self, self, indicies)
		# CLEAN UP EMPTY TOP LEVEL CONTAINERS
		if len(self._xml_asset) == 0:
			assets = self._xml_root.find('asset')
			self._xml_root.remove(assets)
		if len(self._xml_sensor) == 0:
			sensors = self._xml_root.find('sensor')
			self._xml_root.remove(sensors)
		if len(self._xml_actuator) == 0:
			actuators = self._xml_root.find('actuator')
			self._xml_root.remove(actuators)
		# SET INDENTS
		xml.indent(self._xml_root, '\t', 0)
		self._built = True
		self.reset()


	def reset(self):
		"""
		This method resets the Mujoco simulation data and initializes a single first step 
		to make all runtime data accessable.
		"""
		if self._built:
			xml_string     = self.to_xml_string()
			self._mj_model = mujoco.MjModel.from_xml_string(xml_string)
			self._mj_data  = mujoco.MjData(self._mj_model)
			mujoco.mj_forward(self._mj_model, self._mj_data)
		else:
			raise Exception('World can only be reset after it has been built.')
		


	@blue.restrict
	def unbuild(self) -> None:
		"""
		Unbuilds the World to anable alterations. This discards the current Mujoco simulation.
		"""
		if self._built:
			self._built = False
			for tendon in self.descendants['tendons']['descendants']:
				tendon._built = False
			# Reset _built on material/texture assets so they
			# re-emit their XML elements on the next build.
			for thing in self.all:
				asset = getattr(thing, 'asset', None)
				if asset is not None and hasattr(asset, '_built'):
					asset._built = False
					tex = getattr(asset, 'texture', None)
					if tex is not None:
						tex_asset = getattr(tex, 'asset', None)
						if tex_asset is not None and hasattr(tex_asset, '_built'):
							tex_asset._built = False
			if self.texture is not None:
				self.texture.asset._built = False
			self._name_manager.unregister()
			self._gl_context.free()
			del self._mj_model, self._mj_data, self._gl_context
			self.stop_recording()


	@blue.restrict
	def view(self, callback=lambda: None) -> None:
		"""
		Launches an interactive window of the Mujoco simulation.

		Parameters
		----------
		callback : function
			The callback function is called after each timestep. The user can use this 
			function to update World properties during the simulation.
		"""
		self.build()
		xml_string = self.to_xml_string()
#		xml_string = """
#<mujoco model="humanoid">
#    <compiler angle="degree" inertiafromgeom="true"/>
#    <default>
#        <joint armature="1" damping="1" limited="true"/>
#        <geom conaffinity="1" condim="1" contype="1" margin="0.001" material="geom" rgba="0.8 0.6 .4 1"/>
#        <motor ctrllimited="true" ctrlrange="-.4 .4"/>
#    </default>
#    <option integrator="RK4" iterations="50" solver="PGS" timestep="0.003">
#        <!-- <flags solverstat="enable" energy="enable"/>-->
#    </option>
#    <size nkey="5" nuser_geom="1"/>
#    <visual>
#        <map fogend="5" fogstart="3"/>
#    </visual>
#    <asset>
#        <texture builtin="gradient" height="100" rgb1=".4 .5 .6" rgb2="0 0 0" type="skybox" width="100"/>
#        <!-- <texture builtin="gradient" height="100" rgb1="1 1 1" rgb2="0 0 0" type="skybox" width="100"/>-->
#        <texture builtin="flat" height="1278" mark="cross" markrgb="1 1 1" name="texgeom" random="0.01" rgb1="0.8 0.6 0.4" rgb2="0.8 0.6 0.4" type="cube" width="127"/>
#        <texture builtin="checker" height="100" name="texplane" rgb1="0 0 0" rgb2="0.8 0.8 0.8" type="2d" width="100"/>
#        <material name="MatPlane" reflectance="0.5" shininess="1" specular="1" texrepeat="60 60" texture="texplane"/>
#        <material name="geom" texture="texgeom" texuniform="true"/>
#    </asset>
#    <worldbody>
#        <light cutoff="100" diffuse="1 1 1" dir="-0 0 -1.3" directional="true" exponent="1" pos="0 0 1.3" specular=".1 .1 .1"/>
#        <geom condim="3" friction="1 .1 .1" material="MatPlane" name="floor" pos="0 0 0" rgba="0.8 0.9 0.8 1" size="20 20 0.125" type="plane"/>
#        <!-- <geom condim="3" material="MatPlane" name="floor" pos="0 0 0" size="10 10 0.125" type="plane"/>-->
#        <body name="torso" pos="0 0 1.4">
#            <camera name="track" mode="trackcom" pos="0 -4 0" xyaxes="1 0 0 0 0 1"/>
#            <joint armature="0" damping="0" limited="false" name="root" pos="0 0 0" stiffness="0" type="free"/>
#            <geom fromto="0 -.07 0 0 .07 0" name="torso1" size="0.07" type="capsule"/>
#            <geom name="head" pos="0 0 .19" size=".09" type="sphere" user="258"/>
#            <geom fromto="-.01 -.06 -.12 -.01 .06 -.12" name="uwaist" size="0.06" type="capsule"/>
#            <body name="lwaist" pos="-.01 0 -0.260" quat="1.000 0 -0.002 0">
#                <geom fromto="0 -.06 0 0 .06 0" name="lwaist" size="0.06" type="capsule"/>
#                <joint armature="0.02" axis="0 0 1" damping="5" name="abdomen_z" pos="0 0 0.065" range="-45 45" stiffness="20" type="hinge"/>
#                <joint armature="0.02" axis="0 1 0" damping="5" name="abdomen_y" pos="0 0 0.065" range="-75 30" stiffness="10" type="hinge"/>
#                <body name="pelvis" pos="0 0 -0.165" quat="1.000 0 -0.002 0">
#                    <joint armature="0.02" axis="1 0 0" damping="5" name="abdomen_x" pos="0 0 0.1" range="-35 35" stiffness="10" type="hinge"/>
#                    <geom fromto="-.02 -.07 0 -.02 .07 0" name="butt" size="0.09" type="capsule"/>
#                    <body name="right_thigh" pos="0 -0.1 -0.04">
#                        <joint armature="0.01" axis="1 0 0" damping="5" name="right_hip_x" pos="0 0 0" range="-25 5" stiffness="10" type="hinge"/>
#                        <joint armature="0.01" axis="0 0 1" damping="5" name="right_hip_z" pos="0 0 0" range="-60 35" stiffness="10" type="hinge"/>
#                        <joint armature="0.0080" axis="0 1 0" damping="5" name="right_hip_y" pos="0 0 0" range="-110 20" stiffness="20" type="hinge"/>
#                        <geom fromto="0 0 0 0 0.01 -.34" name="right_thigh1" size="0.06" type="capsule"/>
#                        <body name="right_shin" pos="0 0.01 -0.403">
#                            <joint armature="0.0060" axis="0 -1 0" name="right_knee" pos="0 0 .02" range="-160 -2" type="hinge"/>
#                            <geom fromto="0 0 0 0 0 -.3" name="right_shin1" size="0.049" type="capsule"/>
#                            <body name="right_foot" pos="0 0 -0.45">
#                                <geom name="right_foot" pos="0 0 0.1" size="0.075" type="sphere" user="0"/>
#                            </body>
#                        </body>
#                    </body>
#                    <body name="left_thigh" pos="0 0.1 -0.04">
#                        <joint armature="0.01" axis="-1 0 0" damping="5" name="left_hip_x" pos="0 0 0" range="-25 5" stiffness="10" type="hinge"/>
#                        <joint armature="0.01" axis="0 0 -1" damping="5" name="left_hip_z" pos="0 0 0" range="-60 35" stiffness="10" type="hinge"/>
#                        <joint armature="0.01" axis="0 1 0" damping="5" name="left_hip_y" pos="0 0 0" range="-110 20" stiffness="20" type="hinge"/>
#                        <geom fromto="0 0 0 0 -0.01 -.34" name="left_thigh1" size="0.06" type="capsule"/>
#                        <body name="left_shin" pos="0 -0.01 -0.403">
#                            <joint armature="0.0060" axis="0 -1 0" name="left_knee" pos="0 0 .02" range="-160 -2" stiffness="1" type="hinge"/>
#                            <geom fromto="0 0 0 0 0 -.3" name="left_shin1" size="0.049" type="capsule"/>
#                            <body name="left_foot" pos="0 0 -0.45">
#                                <geom name="left_foot" type="sphere" size="0.075" pos="0 0 0.1" user="0" />
#                            </body>
#                        </body>
#                    </body>
#                </body>
#            </body>
#            <body name="right_upper_arm" pos="0 -0.17 0.06">
#                <joint armature="0.0068" axis="2 1 1" name="right_shoulder1" pos="0 0 0" range="-85 60" stiffness="1" type="hinge"/>
#                <joint armature="0.0051" axis="0 -1 1" name="right_shoulder2" pos="0 0 0" range="-85 60" stiffness="1" type="hinge"/>
#                <geom fromto="0 0 0 .16 -.16 -.16" name="right_uarm1" size="0.04 0.16" type="capsule"/>
#                <body name="right_lower_arm" pos=".18 -.18 -.18">
#                    <joint armature="0.0028" axis="0 -1 1" name="right_elbow" pos="0 0 0" range="-90 50" stiffness="0" type="hinge"/>
#                    <geom fromto="0.01 0.01 0.01 .17 .17 .17" name="right_larm" size="0.031" type="capsule"/>
#                    <geom name="right_hand" pos=".18 .18 .18" size="0.04" type="sphere"/>
#                    <camera pos="0 0 0"/>
#                </body>
#            </body>
#            <body name="left_upper_arm" pos="0 0.17 0.06">
#                <joint armature="0.0068" axis="2 -1 1" name="left_shoulder1" pos="0 0 0" range="-60 85" stiffness="1" type="hinge"/>
#                <joint armature="0.0051" axis="0 1 1" name="left_shoulder2" pos="0 0 0" range="-60 85" stiffness="1" type="hinge"/>
#                <geom fromto="0 0 0 .16 .16 -.16" name="left_uarm1" size="0.04 0.16" type="capsule"/>
#                <body name="left_lower_arm" pos=".18 .18 -.18">
#                    <joint armature="0.0028" axis="0 -1 -1" name="left_elbow" pos="0 0 0" range="-90 50" stiffness="0" type="hinge"/>
#                    <geom fromto="0.01 -0.01 0.01 .17 -.17 .17" name="left_larm" size="0.031" type="capsule"/>
#                    <geom name="left_hand" pos=".18 -.18 .18" size="0.04" type="sphere"/>
#                </body>
#            </body>
#        </body>
#    </worldbody>
#    <tendon>
#        <fixed name="left_hipknee">
#            <joint coef="-1" joint="left_hip_y"/>
#            <joint coef="1" joint="left_knee"/>
#        </fixed>
#        <fixed name="right_hipknee">
#            <joint coef="-1" joint="right_hip_y"/>
#            <joint coef="1" joint="right_knee"/>
#        </fixed>
#    </tendon>
#
#    <actuator>
#        <motor gear="100" joint="abdomen_y" name="abdomen_y"/>
#        <motor gear="100" joint="abdomen_z" name="abdomen_z"/>
#        <motor gear="100" joint="abdomen_x" name="abdomen_x"/>
#        <motor gear="100" joint="right_hip_x" name="right_hip_x"/>
#        <motor gear="100" joint="right_hip_z" name="right_hip_z"/>
#        <motor gear="300" joint="right_hip_y" name="right_hip_y"/>
#        <motor gear="200" joint="right_knee" name="right_knee"/>
#        <motor gear="100" joint="left_hip_x" name="left_hip_x"/>
#        <motor gear="100" joint="left_hip_z" name="left_hip_z"/>
#        <motor gear="300" joint="left_hip_y" name="left_hip_y"/>
#        <motor gear="200" joint="left_knee" name="left_knee"/>
#        <motor gear="25" joint="right_shoulder1" name="right_shoulder1"/>
#        <motor gear="25" joint="right_shoulder2" name="right_shoulder2"/>
#        <motor gear="25" joint="right_elbow" name="right_elbow"/>
#        <motor gear="25" joint="left_shoulder1" name="left_shoulder1"/>
#        <motor gear="25" joint="left_shoulder2" name="left_shoulder2"/>
#        <motor gear="25" joint="left_elbow" name="left_elbow"/>
#    </actuator>
#</mujoco>
#"""
#		self._mj_model = mujoco.MjModel.from_xml_string(xml_string)
#		self._mj_data  = mujoco.MjData(self._mj_model)
#		mujoco.mj_forward(self._mj_model, self._mj_data)
		
		print(xml_string)
		glfw.init()
		with mujoco.viewer.launch_passive(self._mj_model, self._mj_data, show_left_ui=True, show_right_ui=True) as viewer:
			self._viewer = viewer
			## TIME
			#start_time = time.time()
			# DEFAULT CAMERA POSITIONING
			viewer.cam.distance = self.size * 4
			viewer.cam.lookat = self.center
			desc = self.descendants['cameras']['descendants']
			while viewer.is_running():
				mujoco.mj_step(self._mj_model, self._mj_data)
				viewer.sync()
				# CALLBACK
				callback()
				#print(viewer.update_hfield.)
				#print(*sorted(viewer.__dir__()), sep='\n')
				# TIME SYNCH
				#model_time = self._mj_data.time
				#run_time   = time.time() - start_time
				#delta      = max(0, model_time - run_time)
				#time.sleep(delta)
			self._viewer = None


	@blue.restrict
	def step(self, 
		 n_steps: int|None       = 1, 
		 seconds: int|float|None = None, 
		 pbar:    bool           = False) -> None:
		"""
		Performs timesteps in the simmulation. Exactly one argument must be given by the user. 
		All cameras that are recording are taking images during the simulation at their respective 
		times.

		Parameters
		----------
		n_steps : int | None
			The number of steps to be taken
		seconds : int | float | None
			The number of seconds to be simulated
		"""
		#if not hasattr(self, '_mj_model') and not hasattr(self, '_mj_data'):
		if not self._built:
			raise Exception('The World must first be build before simulation steps can be performed. Use World.build()!')
		if n_steps is None and seconds is None:
			raise ValueError('Precisely one argument seconds or n_steps must be set, but both were None.')
		if seconds is not None:
			n_steps = int(seconds/self.timestep)
			if n_steps == 0:
				raise ValueError(f'With World.timestep={self.timestep} and seconds={seconds} the number of simulation steps is 0. Please decrease the timestep or increase the duration (seconds).')
		if pbar:
			iterator = tqdm.tqdm(range(n_steps))
		else:
			iterator = range(n_steps)
		for n in iterator:
			
			mujoco.mj_step(self._mj_model, self._mj_data)
			time = self._mj_data.time
			for camera, rec_dict in self._recordings.items():
				fps, last, writer = rec_dict['fps'], rec_dict['last'], rec_dict['writer']
				if last + 1/fps <= time:
					writer.append_data(camera.observation)
					rec_dict['last'] = time
		#for descendants in self.descendants.values():
		#	for descendant in descendants['descendants']:
		#		descendant._clear_step_cache()


	@blue.restrict
	def start_recording(self, 
			    *cameras: list[blue.CameraType], 
			    filename: str, 
			    fps:      int = 60,  
			    **kwargs):
		"""
		This method starts recordings of the cameras passed to it. The rendering is done 
		by the mujoco.Renderer using OpenGL. The user should make sure, that their OpenGL 
		instance is properly functioning. The video files are written to the ``recordings`` 
		directory next to the file instantiating the World. The format is infered from the 
		file extension in the ``filename`` using imageio. To set additional parameters for 
		the video formatting in imageio the user can set the ``**kwargs`` which are passed 
		to the imageio writer.

		Example
		-------
		If the user would like to record a gif image that loops forever (instead of 
		the default case of a single loop) the user can set the argument ``loop`` to 
		zero, indicating an infinitely looping gif in the imageio writer:

		>>> world.start_recording(camera, filename='my.gif', loop=0)

		Now the camera has started recording, but there is nothing yet happening in the 
		simulation. Using :meth:`World.step` the user can simulate the passage of time with 
		the recording cameras writing images to their respective video files at correpsonding 
		moments (as determined by ``fps`` and :attr:`World.timestep`). Additionally the user 
		can apply actions in between timesteps via an :class:`Agent <blueprints.agent.Agent>` 
		or through :class:`Actuators <blueprints.actuators.BaseActuator>` directly.

		For example, the following lines demonstrate such a simulation loop with some number 
		of steps in MDP space determined by ``time_steps``, some ``skip_frames``, a blueprints 
		:class:`Agent <blueprints.agent.Agent>` ``agent`` used to interface with the :class:`Sensors <blueprints.sensors.BaseSensor>` 
		and :class:`Actuators <blueprints.actuators.BaseActuator>` and an RL-Agent ``policy``
		
		>>> for t in range(time_steps):
		>>> 	    world.step(n_steps=skip_frames)
		>>> 	    agent.force = policy(agent.observations)

		Finally the recordings are stopped using :meth:`stop_recording` to save the recorded 
		videos to a file.

		>>> world.stop_recording()

		The final video of the users agent can then be found in ``'recordings/camera_my.gif'``

		Parameters
		----------
		*cameras : list [ blue.CameraType ]
			A list of :class:`Cameras <blueprints.camera.Camera>`
		filename : str
			The name (with a proper file extension indicating the format) of the video.
			If multiple cameras are passed, multiple videos will be created with the name 
			of each camera prefixed to the filename.
		fps : int
			The frames per second
		**kwargs : dict
			Additional keyword arguments that are passed to imageio to write the video.
		"""
		# VALUE CHECKS
		if not all(camera in self.all.cameras for camera in cameras):
			raise ValueError(f'Not all cameras given are attached within the kinematic hierarchy of the World.')
		if not self._built:
			raise Exception('World must first be build before recordings can start.')
		# MAKE DIRECTORIES
		path = f'{self._path}/recordings'
		if not os.path.exists(path):
			os.mkdir(path)
		for camera in cameras:
			writer = imageio.get_writer(f'{path}/{camera.name}_{filename}', fps=fps, **kwargs)
			self._recordings[camera] = {'fps': fps, 'writer': writer, 'last': 0.}


	@blue.restrict
	def stop_recording(self, 
			   *cameras: list[blue.CameraType]):
		"""
		This method stops recording for all the cameras passed. If no camera is passed, all recording cameras will be stopped.

		Parameters
		----------
		cameras : list [ blue.CameraType ]
			The cameras to be stopped.
		"""
		# VALUE CHECKS
		if not all(camera in self.all.cameras for camera in cameras):
			raise ValueError(f'Not all cameras given are attached within the kinematic hierarchy of the World.')
		if not cameras:
			for rec_dict in self._recordings.values():
				rec_dict['writer'].close()
			self._recordings = dict()
		else:
			for camera in cameras:
				self._recordings[camera]['writer'].close()
				del self._recordings[camera]


	@blue.restrict
	def to_xml_string(self) -> str:
		"""
		If not build, this method will first build the world, retrieve the XML string and then unbuild the world again. 
		If the user calls this method repeatedly the world should probably be build first.

		Returns
		-------
		str
			The xml string representing the xml model used to launch Mujoco.
		"""
		built = self._built
		if not built:
			self.build()
		string = xml.tostring(self._xml_root, encoding='unicode')
		if not built:
			self.unbuild()
		return string


	@blue.restrict
	@classmethod
	def from_xml_string(cls, string: str) -> blue.WorldType:
		"""
		This method reconstructs a World from a Mujoco xml string. It currently only supports reading 
		models, that exclusively consist of blueprints supported features. In a future version it will 
		be extended to be compatible to all Mujoco models.
		
		Parameters
		----------
		string : str
			The xml description of the Mujoco model.
		
		Returns
		-------
		blue.WorldType
			The reconstructed World.
		"""
		init_args     = dict()
		caches        = defaultdict(dict)
		assets        = defaultdict(dict)
		materials     = dict()
		sensors       = defaultdict(lambda: defaultdict(list))
		actuators     = defaultdict(lambda: defaultdict(list))
		ref_actuators = defaultdict(lambda: defaultdict(list))
		TEXTURE_CLASSES = {'2d':     blue.texture.Plane,
				   'cube':   blue.texture.Box,
				   'skybox': blue.texture.Skybox}
		xml_tree      = xml.fromstring(string)
		xml_compiler  = xml_tree.find('compiler')
		xml_assets    = xml_tree.find('asset')
		xml_sensors   = xml_tree.find('sensor')
		xml_actuators = xml_tree.find('actuator')
		xml_worldbody = xml_tree.find('worldbody')
		xml_option    = xml_tree.find('option')
		# COLLECTING ARGS
		name = xml_tree.get('model')
		if name is not None:
			init_args['name'] = name
		if xml_compiler is not None:
			angle = xml_compiler.get('angle')
			if angle is not None:
				init_args['angle'] = angle
			autolimits = xml_compiler.get('autolimits')
			if autolimits is not None:
				init_args['autolimits'] = autolimits == 'true'
		if xml_option is not None:
			init_args['viscosity']  = float(xml_option.get('viscosity')) if xml_option.get('viscosity') is not None else 0.
			init_args['integrator'] = xml_option.get('integrator')
			timestep = xml_option.get('timestep')
			if timestep is not None:
				init_args['timestep'] = float(timestep)
			cone = xml_option.get('cone')
			if cone is not None:
				init_args['cone'] = cone
			xml_flag = xml_option.find('flag')
			if xml_flag is not None:
				init_args['gravity'] = xml_flag.get('gravity') == 'enable'
				init_args['contact'] = xml_flag.get('contact') == 'enable'
		if xml_sensors is not None:
			for sensor in xml_sensors:
				sensor_obj    = blue.REGISTER._get_thing(sensor)
				sensor_parent = sensor_obj._PARENT_TYPE
				sensors[sensor_parent][sensor.get(sensor_parent)].append(sensor_obj)
		if xml_assets is not None:
			# XML attr name → Texture constructor param name
			TEX_ATTR_MAP = {'file':      'filename',     'rgb1':       'color_1',
					'rgb2':      'color_2',      'markrgb':    'color_mark',
					'gridsize':  'grid_size',    'gridlayout': 'grid_layout',
					'hflip':     'h_flip',       'vflip':      'v_flip',
					'nchannel':  'n_channel',    'content_type': 'content',
					'fileright': 'filename_right',  'fileleft':  'filename_left',
					'fileup':    'filename_up',     'filedown':  'filename_down',
					'filefront': 'filename_front',  'fileback':  'filename_back'}
			TEX_FLOAT  = {'random'}
			TEX_INT    = {'width', 'height', 'n_channel', 'nchannel'}
			TEX_BOOL   = {'hflip', 'vflip', 'h_flip', 'v_flip'}
			TEX_ARRAY  = {'rgb1', 'rgb2', 'markrgb', 'gridsize'}
			TEX_SKIP   = {'type', 'name'}
			# First pass: textures and cache-based assets (mesh, hfield)
			for asset in xml_assets:
				asset_type = asset.tag
				if asset_type == 'texture':
					tex_type   = asset.get('type', '2d')
					tex_class  = TEXTURE_CLASSES[tex_type]
					tex_kwargs = {}
					for k, v in asset.items():
						if k in TEX_SKIP:
							continue
						param = TEX_ATTR_MAP.get(k, k)
						if k in TEX_FLOAT:
							v = float(v)
						elif k in TEX_INT:
							v = int(v)
						elif k in TEX_BOOL:
							v = v == 'true'
						elif k in TEX_ARRAY:
							v = [float(x) for x in v.split()]
						tex_kwargs[param] = v
					tex_name = asset.get('name')
					tex_obj = tex_class(name=tex_name, **tex_kwargs)
					tex_obj.asset._name = tex_name
					assets['texture'][tex_name] = tex_obj
				elif asset_type in blue.REGISTER.CACHE_THINGS:
					asset_name = asset.get('name')
					if asset_name in caches[asset_type]:
						cache = caches[asset_type][asset_name]
					else:
						cache_class = blue.REGISTER.CACHE_THINGS[asset_type]
						cache = cache_class._from_xml_element(asset)
					asset_class = blue.REGISTER._get_thing_class(asset)
					asset_obj   = asset_class._from_xml_element(xml_element=asset,
											cache=cache)
					assets[asset_type][asset.get('name')] = asset_obj
			# Second pass: materials (reference textures by name)
			MAT_FLOAT = {'emission', 'specular', 'shininess', 'reflectance',
				      'metallic', 'roughness'}
			MAT_BOOL  = {'texuniform'}
			MAT_ARRAY = {'texrepeat', 'rgba'}
			MAT_SKIP  = {'name', 'texture'}
			for asset in xml_assets:
				if asset.tag == 'material':
					tex_name   = asset.get('texture')
					texture    = assets['texture'].get(tex_name) if tex_name else None
					mat_kwargs = {}
					for k, v in asset.items():
						if k in MAT_SKIP:
							continue
						if k == 'rgba':
							mat_kwargs['color'] = [float(x) for x in v.split()]
						elif k in MAT_FLOAT:
							mat_kwargs[k] = float(v)
						elif k in MAT_BOOL:
							mat_kwargs[k] = v == 'true'
						elif k in MAT_ARRAY:
							mat_kwargs[k] = [float(x) for x in v.split()]
						else:
							mat_kwargs[k] = v
					mat_name = asset.get('name')
					mat_obj = blue.Material(texture=texture,
								name=mat_name,
								**mat_kwargs)
					# Preserve the original asset name so the
					# round-tripped XML uses the same identifier.
					mat_obj.asset._name = mat_name
					materials[mat_name] = mat_obj
		if xml_actuators is not None:
			for actuator in xml_actuators:
				actuator_obj  = blue.REGISTER._get_thing(actuator)
				actuator_obj._IGNORE_CHECKS = True
				actuator_name = actuator.get('name')
				# Attach sensors that reference this actuator
				if actuator_name in sensors['actuator']:
					actuator_obj.attach(*sensors['actuator'][actuator_name], copy=False)
				# Find the parent reference attribute (joint, site, body, or jointinparent)
				parent_name = None
				parent_key  = None
				for attr in ('joint', 'jointinparent', 'site', 'body'):
					parent_name = actuator.get(attr)
					if parent_name is not None:
						parent_key = 'joint' if attr == 'jointinparent' else attr
						break
				if parent_key is not None:
					actuators[parent_key][parent_name].append(actuator_obj)
				# Handle secondary references (refsite)
				for ref_attr, ref_type in actuator_obj._OTHER_REFERENCES.items():
					ref_name = actuator.get(ref_attr)
					if ref_name is not None:
						ref_actuators[ref_type][ref_name].append(actuator_obj)
		# Assign skybox texture to init_args if present
		for tex_obj in assets.get('texture', {}).values():
			if isinstance(tex_obj, blue.texture.Skybox):
				init_args['texture'] = tex_obj
				break
		# INIT WORLD
		world = object.__new__(cls)
		world.__init__(**init_args)
		for xml_element in xml_worldbody:
			cls._build_from_xml(parent=world,
						xml_element=xml_element,
						assets=assets,
						materials=materials,
						sensors=sensors,
						actuators=actuators,
						ref_actuators=ref_actuators)
		for tag_actuators in actuators.values():
			for name_parent in tag_actuators.values():
				for actuator in name_parent:
					actuator._IGNORE_CHECKS = False
		# Resolve camera/light target references by name
		all_things = {t.name: t for t in world.all}
		for thing in world.all:
			if isinstance(thing, (blue.Camera, blue.Light)):
				target_name = getattr(thing, '_target_name', None)
				if target_name is not None and target_name in all_things:
					thing.target = all_things[target_name]
		return world


	@property
	def model(self):
		"""
		If a feature is not accessable through blueprints you can 
		access the Mujoco model directly.

		Returns
		-------
		mj.Model
		"""
		if self._built:
			return self._mj_model
		else:
			raise Exception('Mujoco model is only accessable after the World has been built.')

	@property
	def data(self):
		"""
		If a feature is not accessable through blueprints you can 
		access the Mujoco data directly.

		Returns
		-------
		mj.Data
		"""
		if self._built:
			return self._mj_data
		else:
			raise Exception('Mujoco data is only accessable after the World has been built.')
	


	@blue.restrict
	@classmethod
	def _build_from_xml(cls,
			    parent,
			    xml_element,
			    assets:        dict,
			    materials:     dict,
			    sensors:       dict,
			    actuators:     dict,
			    ref_actuators: dict) -> blue.ThingType:
		"""
		This method builds the world from the xml Element.

		Parameters
		----------
		parent : WorldType
			The World object to which the reconstructed Things are attached.
		xml_element : TYPE
			The xml element from which the Things used for reconstruction are taken.
		assets : dict
			A dictionary of assets.
		materials : dict
			A dictionary of materials.
		sensors : dict
			A dictionary of sensors.
		actuators : dict
			A dictionary of actuators.
		ref_actuators : dict
			A dictionary of ref_actuators.
		"""
		# CONSTRUCT ELEMENT
		xml_type  = xml_element.get('type')
		xml_tag   = xml_element.tag
		xml_name  = xml_element.get('name', '')
		# Detect Agent bodies by AGENT: prefix
		if xml_tag == 'body' and xml_name.startswith('AGENT:'):
			obj_class = blue.Agent
		else:
			obj_class = blue.REGISTER._get_thing_class(xml_element)
		xml_args  = dict()
		if xml_type in assets:
			xml_name = xml_element.get(xml_type)
			xml_args['asset'] = assets[xml_type][xml_name]
		mat_name = xml_element.get('material')
		if mat_name is not None and mat_name in materials:
			xml_args['material'] = materials[mat_name]
		if xml_tag in sensors:
			xml_name = xml_element.get('name')
			xml_args['sensors'] = sensors[xml_tag][xml_name]
		if xml_tag in actuators:
			xml_name = xml_element.get('name')
			xml_args['actuators'] = actuators[xml_tag][xml_name]
		if xml_tag in ref_actuators:
			xml_name = xml_element.get('name')
			xml_args['ref_actuators'] = ref_actuators[xml_tag][xml_name]
		obj = obj_class._from_xml_element(xml_element=xml_element, **xml_args)
		# CONSTRUCT CHILDREN
		for xml_child in xml_element:
			cls._build_from_xml(parent=obj,
						xml_element=xml_child,
						assets=assets,
						materials=materials,
						sensors=sensors,
						actuators=actuators,
						ref_actuators=ref_actuators)
		parent.attach(obj, copy=False)

	# PROPERTIES

	@property
	def size(self):
		"""
		The size is computed to sclae the FreeCamera of the viewer (see :meth:`World.view`).

		Returns
		-------
		np.ndarray
			The size of the World ignoring background Things.
		"""
		built = self._built
		if not built:
			self.build()
		size = self._size
		if not built:
			self.unbuild()
		return size


	@property
	def center(self):
		"""
		The center is computed to position the FreeCamera of the viewer (see :meth:`World.view`).

		Returns
		-------
		np.ndarray
			The center of the World ignoring background Things.
		"""
		built = self._built
		if not built:
			self.build()
		center = self._center
		if not built:
			self.unbuild()
		return center


	@property
	def texture(self) -> blue.SkyboxTextureType:
		"""
		Only a :class:`SkyboxTexture <blueprints.texture.Skybox>` are valid for the World.

		.. note::
			At the moment, Mujoco is only rendering SkyboxTextures of the World, if also another 
			texture is taken by a material refferenced by a Thing within the World.

		Returns
		-------
		blue.SkyboxTextureType
			The Skybox Texture which is applied to the background of the World.
		"""
		return self._texture


	@texture.setter
	@blue.restrict
	def texture(self, texture: blue.SkyboxTextureType|None) -> None:
		if texture is not None:
			self._texture = texture.copy()
			self._texture._parent = self
		else:
			self._texture = texture



import numpy as np
import blueprints as blue



class BaseTube(blue.TubeType):

	"""
	BaseTube is a base class for sites and geoms that can be defined by two points forming a tube like object.
	"""
	
	@blue.restrict
	@classmethod
	def from_points(cls, 
			head:   np.ndarray|list[int|float], 
			tail:   np.ndarray|list[int|float], 
			radius: int|float = 1, 
			**kwargs) -> blue.TubeType:
		"""
		Constructs a Tube like object from point head to point tail.
		Euler angles are set such that the resulting orientation
		has pitch and yaw but no roll. Therefore the floors horizon
		is always parallel to the orientation.
		
		Args:
			head (np.ndarray | list[int | float]): Description
			tail (np.ndarray | list[int | float]): Description
			radius (int | float, optional): Description
			**kwargs: Description
		
		Returns:
			blue.Thing: Description
		"""
		head   = np.array(head)
		tail   = np.array(tail)
		path   =  tail - head
		pos    = (head + tail)/2
		length = np.linalg.norm(path)
		# RECONSTRUCT REFERENCE FRAME
		R_z    = path/length
		R_z1, R_z2, R_z3 = R_z
		if R_z1 == R_z2 == 0:
			alpha = 0 if R_z3 > 0 else blue.geometry.PI
			beta  = 0
			gamma = 0
		else:
			R_x = blue.geometry.Vector.normalize(np.array([R_z2, -R_z1, 0]))
			R_y = np.cross(R_z, R_x)
			R   = np.stack([R_x, R_y, R_z], axis=1)
			# RECONSTRUCT ANGLES
			alpha, beta, gamma = blue.geometry.Rotation.reference_frame_to_euler(R)
		kwargs['pos']    = pos
		kwargs['alpha']  = alpha
		kwargs['beta']   = beta
		kwargs['gamma']  = gamma
		if issubclass(cls, (blue.BoxGeomType, blue.BoxSiteType)):
			kwargs['x_length'] = radius
			kwargs['y_length'] = radius
			kwargs['z_length'] = length
		else:
			kwargs['length'] = length
			kwargs['radius'] = radius
		return cls(**kwargs)


	@property
	def head(self) -> np.ndarray:
		"""
		Returns the head position of the tube like object. The positions are derived even if the object was not constructed with the from points method.
		
		Returns:
			np.ndarray: The head position.
		"""
		R = self.rotation_matrix
		length = self.z_length if isinstance(self, (blue.BoxGeomType, blue.BoxSiteType)) else self.length
		Z = blue.geometry.Rotation.Z * length / 2
		return self.pos - R @ Z


	@property
	def tail(self) -> np.ndarray:
		"""
		Returns the tail position of the tube like object. The positions are derived even if the object was not constructed with the from points method.
		
		Returns:
			np.ndarray: The tail position.
		"""
		R = self.rotation_matrix
		length = self.z_length if isinstance(self, (blue.BoxGeomType, blue.BoxSiteType)) else self.length
		Z = blue.geometry.Rotation.Z * length / 2
		return self.pos + R @ Z


	@head.setter
	@blue.restrict
	def head(self, head: np.ndarray|list[int|float]) -> None:
		"""
		Setter for the head parameter. The position and orientation of the object is adjusted accordingly.
		
		Args:
			head (np.ndarray | list[int | float]): Description
		
		Returns:
			blue.Thing: The new Geom or Site with the adjusted position and orientation.
		"""
		tube        = self.from_points(head, self.tail, **self._blueprint_specs())
		self.pos    = tube.pos
		self.euler  = tube.euler
		self.length = tube.length


	@tail.setter
	@blue.restrict
	def tail(self, tail: np.ndarray|list[int|float]) -> None:
		"""
		Setter for the tail parameter. The position and orientation of the object is adjusted accordingly.
		
		Args:
			tail (np.ndarray | list[int | float]): Description
		
		Returns:
			blue.Thing: The new Geom or Site with the adjusted position and orientation.
		"""
		tube        = self.from_points(self.head, tail, **self._blueprint_specs())
		self.pos    = tube.pos
		self.euler  = tube.euler
		self.length = tube.length



"""
.. note::
	If you found yourself here not being a dev, you should probably :mod:`skip this section <blueprints.thing.moveable>` or 
	look up the correspinging docs page of the Thing you were looking after for which this module only 
	implements its corresponding mirror types. A link to the docs page is to be found in the Types doc 
	string.
"""

import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict



class CyclicalThing(blue.CyclicalThingType, base.BaseThing):
	"""
	If a Thing has a reference, that potentially points above its parent in the hierarchy of the 
	kinematic tree it becomes a kinematic graph, loosing important properties that are relied on for 
	copying. This class is used to enable migration of attributes that might induce a cycle and to 
	mitigated the introduction of a cycle. The :class:`CyclicalThing` should only be properly attached 
	to one parent, keeping the other references back into the kinematic tree as separate non children 
	attributes (making them improper parents). If a reference of a parent (proper and improper) is 
	performing a copy it should not assign a copy of this instance to its attribute, but instead use 
	the :meth:`_migrate` method. This is used to ensure that this instance is not copied once for each 
	parent.

	Additionally, attributes are restricted from being set to another Thing, if it is not part of the 
	same kinematic tree. This is necessary to prevent exuberant duplications of :math:`c^p` with :math:`c` as the 
	number of copies and :math:`p` as the number of parents (proper and improper).

	If a modification of the kinematic tree results in a parent reference that points outside the 
	kinematic tree, the reference is reset to None.

	The class definition must define two sets ``_PARENT_REFERENCES`` (proper) and ``_OTHER_REFERENCES`` 
	(improper) as class attributes containing the attribute names under which the parents are assigned 
	to the Thing.
	"""
	def __init__(self, **kwargs):
		"""
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to `super().__init__`
		"""
		assert hasattr(self, '_PARENT_REFERENCE')
		assert hasattr(self, '_OTHER_REFERENCES')
		self._MIGRATIONS    = {}
		self._MIGRATED      = False
		self._IGNORE_CHECKS = False
		super().__init__(**kwargs)


	@property
	def _MIGRATION_DONE(self) -> bool:
		"""
		Returns
		-------
		bool
			Indicates if all references have been migrated.
		"""
		return 'parent' in self._MIGRATIONS.keys() and all(ref in self._MIGRATIONS.keys() for ref in self._OTHER_REFERENCES.keys())


	def _start_migration(self, attr, ref, **kwargs):
		"""
		This method creates a singular copy and assigns all previously gather migrations to it.
		Afterwards all migrations are automatically performed as attribute setting on the copy.
		
		Parameters
		----------
		attr : str
			Name of the attribute.
		ref : ThingType
			The Thing that the attribute is referencing (must be proper parent).
		**kwargs
			Keywords arguments are passed to ```copy`` and will be assigned to the new copy.
		
		Returns
		-------
		ThingType
			The returned value is the copy, possibly not fully constructed, since other improper 
			parents in the kinematic tree might still have to be migrated.
		"""
		self._COPY = self.copy(**kwargs)
		self._COPY._IGNORE_CHECKS = True
		self._MIGRATED = True
		ref.attach(self._COPY, copy=False)
		for attr, ref in self._MIGRATIONS.items():
			self._COPY.__setattr__(attr, ref)
		self._MIGRATIONS[attr] = ref
		self._COPY._IGNORE_CHECKS = False


	def _finalize_migration(self):
		"""
		This method finalizes the migration, after the last reference was assigned.
		"""
		self._COPY._IGNORE_CHECKS = False
		self._COPY       = None
		self._MIGRATED   = False
		self._MIGRATIONS = {}


	def _migrate(self, attr: str, ref: blue.ThingType, ignore_checks=False):
		"""
		This method is used to avoid unintended copies by ensuring, that when this Thing is copied 
		as part of a deep copy of the kinematic tree, only one copy of this instance is generated and 
		all other parents (proper and improper) get assigned just this one copy. Generally parents copy 
		all their children when they are copied, which leads to exuberant duplicate in the case of 
		multiple parents.

		This method is used to differentiate between pre and post copy attribute assignment of 
		proper and improper parents. If an improper parent calls for migration before the proper 
		parent, the improper migration gets saved for later. If the proper parent calls for migration 
		a copy is constructed and all previous migrations are performed. Afterwards all migrations 
		result in an attribute assignment.
		
		Parameters
		----------
		attr : str
			Name of the parent attribute
		ref : blue.ThingType
			The reference to the new parent.
		ignore_checks : bool, optional
			If set the restriction that improper parent attributes are allowed to be assigned to other 
			nodes in the kinematic tree is lifted for the duration of this migration.
		
		Raises
		------
		Exception
			If a reference gets migrated multiple times an error is raised.
		"""
		if attr == 'parent':#in self._PARENT_REFERENCES:
			if self._MIGRATED:
				self._finalize_migration()
			self._start_migration(attr, ref)
		elif self._MIGRATED:
			if attr in self._MIGRATIONS:
				raise Exception(f'The attribute {attr} has already been migrated!')
			self._MIGRATIONS[attr] = ref
			# MIGRATION STEP
			temp = self._COPY._IGNORE_CHECKS
			self._COPY._IGNORE_CHECKS = ignore_checks
			self._COPY.__setattr__(attr, ref)
			self._COPY._IGNORE_CHECKS = temp
		elif attr in self._OTHER_REFERENCES:
			self._MIGRATIONS[attr] = ref
		#else:
		#	if attr == 'parent':#in self._PARENT_REFERENCES:
		#		if self._MIGRATED:
		#			raise Exception(f'CyclicalThings are only allowed to be migrated from a parent once. {repr(self)} was already migrated by its parent {repr(self.parent)}')
		#		self._start_migration(attr, ref)
		#	elif attr in self._OTHER_REFERENCES:
		#		self._MIGRATIONS[attr] = ref
		if self._MIGRATION_DONE:
			self._finalize_migration()


	def _decouple(self):
		"""
		This method ensures that parent attribute pointing outside the kinematic tree, the 
		attribute gets reset to None.
		"""
		for attr in self._OTHER_REFERENCES:
			ref = self.__getattribute__(attr)
			if self._disjunct(self.parent, ref) and ref is not None:
				ref.detach(self)
				self.__setattr__(attr, None)


	@staticmethod
	def _disjunct(thing_a, thing_b):
		"""
		This method checks, whether two things belong to the same kinematic tree.
		
		Parameters
		----------
		thing_a : ThingType
			One thing_a that is checked on whether it shares a root with thing_b.
		thing_b : ThingType
			One thing_b that is checked on whether it shares a root with thing_a.
		
		Returns
		-------
		bool
			Indicates if the two Things share a kinematic tree.
		"""
		return thing_a is None or thing_b is None or thing_a.root is not thing_b.root


	def __setattr__(self, attr, ref) -> None:
		"""
		This method redirects to various other ``__setattr__`` implementations.
		
		Parameters
		----------
		attr : str
			Name of the attribute
		ref : object
			Value of the attribute
		
		Raises
		------
		ValueError
			If the attribute cannot be set an error is raised.
		"""
		if attr == 'parent':
			if self.parent is not None:
				self.parent.detach(self)
			super().__setattr__('parent', ref)
			if not self._IGNORE_CHECKS:
				self._decouple()
		elif attr in self._OTHER_REFERENCES:
			if self._IGNORE_CHECKS or not self._disjunct(self.parent, ref) or ref is None:
				getattr(self.__class__, attr).fset(self, ref)
			else:
				raise ValueError(f'Setting {attr} is only allowed for references that share the same root.')
		else:
			super().__setattr__(attr, ref)



from setuptools import setup, find_packages

with open('README.md', 'r') as file:
	long_description = file.read()

setup(name='mujoco-blueprints', 
      version='0.1.1', 
      packages=find_packages(), 
      author='Mortimer von Chappuis, Yash Saraswat', 
      install_requires=['numpy>=2.0.0', 
      			'imageio>=2.0.0', 
      			'mujoco>=3.3.2', 
      			'tqdm>=4.0.0'], 
      description='Blueprints is a Mujoco interface providing many graph manipulation operations and easy runtime data access.', 
      long_description=long_description, 
      long_description_content_type='text/markdown', 
      url='https://mujoco-blueprints.readthedocs.io/en/latest/index.html', 
      download_url='https://github.com/mortimervonchappuis/mujoco_blueprints')


import blueprints as blue



class Register:

	"""
	The Register is used to assign Thing IDs and to amalgamate all Thing classes for object recostruction from an xml element.
	
	Attributes
	----------
	ACTUATOR_THINGS : dict
		A dictionary containing the respective Actuator classes with their xml tags as keys.
	ASSET_THINGS : dict
		A dictionary containing the respective Asset classes with their xml type attributes as keys.
	assets : list
		A list of all assets that have been created.
	BASIC_THINGS : dict
		A dictionary containing basic Thing classes with their xml tags as keys.
	CACHE_THINGS : dict
		A dictionary containing the respective Cache classes with their xml type attributes as keys.
	caches : list
		A list of all caches that have been created
	DERIVED_THINGS : dict
		A dictionary containing dictionaries of Things for which the specific class is derived from the xml type attribute.
	GEOM_THINGS : dict
		A dictionary containing the respective Geom classes with their xml type attributes as keys.
	JOINT_THINGS : dict
		A dictionary containing the respective Joint classes with their xml type attributes as keys.
	SENSOR_THINGS : dict
		A dictionary containing the respective Sensor classes with their xml tags as keys.
	SITE_THINGS : dict
		A dictionary containing the respective Actuator classes with their xml type attributes as keys.
	"""
	
	BASIC_THINGS    = {'body':   blue.Body, 
			   'camera': blue.Camera, 
			   'light':  blue.Light}
	GEOM_THINGS     = blue.geoms.GEOM_THINGS
	SITE_THINGS     = blue.sites.SITE_THINGS
	JOINT_THINGS    = blue.joints.JOINT_THINGS
	ASSET_THINGS    = blue.assets.ASSET_THINGS
	SENSOR_THINGS   = blue.sensors.SENSOR_THINGS
	ACTUATOR_THINGS = blue.actuators.ACTUATOR_THINGS
	CACHE_THINGS    = blue.cache.CACHE_THINGS
	DERIVED_THINGS  = {'geom':  GEOM_THINGS, 
			   'site':  SITE_THINGS, 
			   'joint': JOINT_THINGS}

	def __init__(self):
		"""
		The Register is used twofold, to keep track of objects and to assign blueprint Thing classes to xml tags.
		"""
		self.__ID = 0
		self.assets = []
		self.caches = []
		self.copy_root = None


	def get_ID(self):
		"""
		This method creates a new ID on each call that can be assigned to a newly initialized Thing. The internal ID counter is incremented afterwards.
		
		Returns
		-------
		int
			A unique ID for a freshly initialized object.
		"""
		ID = self.__ID
		self.__ID += 1
		return ID


	@classmethod
	def _get_thing_class(cls, 
			     xml_element):
		"""
		This method finds the Thing class for a given xml_element.
		
		Parameters
		----------
		xml_element : TYPE
			Description
		
		Returns
		-------
		TYPE
			Description
		"""
		element_tag = xml_element.tag
		if element_tag in cls.BASIC_THINGS:
			obj_class = cls.BASIC_THINGS[element_tag]
		elif element_tag in cls.DERIVED_THINGS:
			DERIVED_CLASSES = cls.DERIVED_THINGS[element_tag]
			element_type = xml_element.get('type')
			if element_type in DERIVED_CLASSES:
				obj_class = DERIVED_CLASSES[element_type]
		elif element_tag in cls.ASSET_THINGS:
			obj_class = cls.ASSET_THINGS[element_tag]
		elif element_tag in cls.SENSOR_THINGS:
			obj_class = cls.SENSOR_THINGS[element_tag]
		elif element_tag in cls.ACTUATOR_THINGS:
			obj_class = cls.ACTUATOR_THINGS[element_tag]
		return obj_class


	@classmethod
	def _get_thing(cls, 
		       xml_element):
		"""
		This method constructs a Thing from an xml element.
		
		Parameters
		----------
		xml_element : xml.etree.ElementTree.Element
			An element from the xml tree paresed with xml.etree.ElementTree.
		
		Returns
		-------
		ThingType
			A thing constructed from the xml_element.
		"""
		thing_class = cls._get_thing_class(xml_element)
		return thing_class._from_xml_element(xml_element)



import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict



class MoveableThing(blue.MoveableThingType, base.BaseThing):
	"""
	This class enables Things to have a position in space and an orientation specified 
	using (improper) euler angles. To place the same Thing in multiple positions and 
	orientations four shortcut methods can be used to construct copies of the Thing on 
	the fly. Those method manipulate position and orientation and this manipulation is 
	either relative to the current position/orientation or absolute, overwriting the current
	position/orientation.


	.. code-block:: python
		:caption: Relative manipulation:

		>>> thing = MoveableThing(pos=[0, 0, 1], alpha=0.8)
		>>> shifted_thing = thing.shift([2, 0, 0])
		>>> shifted_thing.pos
		array([2., 0., 1.], dtype=float32)
		>>> rotated_thing = thing.rotate(beta=0.3)
		>>> rotated_thing.euler
		array([0.8, 0.3, 0. ], dtype=float32)

	.. code-block:: python
		:caption: Absolute manipulation:

		>>> located_thing = thing.locate([2, 0, 0])
		>>> located_thing.pos
		array([2., 0., 0.], dtype=float32)
		>>> oriented_thing = thing.align(beta=0.3)
		>>> oriented_thing.euler
		array([0. , 0.3, 0. ], dtype=float32)

	.. code-block:: python
		:caption: Size manipulation:

		>>> scaled_thing = thing.scaled(3)
		>>> scaled_thing.pos
		array([0., 0., 3.], dtype=float32)
		>>> scaled_thing.size
		array([1.5, 1.5, 1.5], dtype=float32)
	

	For :class:`MoveableThing` instances in a kinematic tree, position and orientation are 
	locally defined, describing the relative change in position and rotation to their parents.
	If one wants to create relocated copies using :meth:`shift` or :meth:`locate` for global 
	coordinates the ``globally`` argument can be set on call.

	Relocation shortcuts also take other :class:`MoveableThing` instances as arguments. This 
	results in a change in position by the arguments position.

	>>> thing_a = MoveableThing(pos=[0, 0, 1])
	>>> thing_b = MoveableThing(pos=[0, 1, 0])
	>>> thing_c = thing_a.shift(pos=thing_b)
	>>> thing_c.pos
	array([0., 1., 1.], dtype=float32)
	"""
	@restrict
	def __init__(self, 
		     pos:   list[int|float]|np.ndarray|None = [0., 0., 0.], 
		     alpha: int|float|None = 0., 
		     beta:  int|float|None = 0., 
		     gamma: int|float|None = 0., 
		     name:  str|None       = None, 
		     x:     int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:     int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:     int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:  list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		MoveableThing implements multiple Thing location and orientation in (improper) 
		euler angles in radians.
		
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		name : str | None, optional
			The user specified name for the Thing.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		#if pos is not None and any(arg is not None for arg in [x, y, z]):
		#	raise ValueError(f'Only argument (pos) or arguments (x, y, z) can be set. Method call got ({', '.join(name for name, val in [('pos', pos), ('x', x), ('y', y), ('z', z)] if val is not None)}).')
		if pos is not None:
			x = x if x is not None else pos[0]
			y = y if y is not None else pos[1]
			z = z if z is not None else pos[2]
		else:
			x = x if x is not None else 0.
			y = y if y is not None else 0.
			z = z if z is not None else 0.
		pos = np.array([x, y, z], dtype=np.float32)
		#if pos is None:
		#	x = float(x) if x is not None else 0.
		#	y = float(y) if y is not None else 0.
		#	z = float(z) if z is not None else 0.
		#	pos = np.array([x, y, z], dtype=np.float32)
		self.pos    = pos
		if quat is not None:
			R, I, J, K = quat[0], quat[1], quat[2], quat[3]
			alpha, beta, gamma = blue.Rotation.quat_to_euler(R, I, J, K)
		self.alpha  = alpha
		self.beta   = beta
		self.gamma  = gamma
		super().__init__(name=name, **kwargs)


	def view(self) -> None:
		"""
		This method creates a throw away :class:`World` <blueprints.world.World>`, 
		attaches itself to it and uses its viewer.

		Returns
		-------
		None
		"""
		world = blue.World()
		#sky = blue.texture.Skybox(builtin='flat', 
		#	  color_1='white')
		#world.texture = sky
		world.attach(self, copy=True)
		world.view()
		world.unbuild()


	def lattice(self, 
		    directions:  list[float|int]|np.ndarray|list[list[int|float]]|list[np.ndarray], 
		    repetitions: int|list[int]|np.ndarray, 
		    name:        str|None = None) -> None:
		"""
		This method constructs a :class:`Lattice <blueprints.utils.lattice.Lattice>` with the MoveableThing as 
		its components.
		
		Parameters
		----------
		directions : list [ float | int ] | np.ndarray | list [ list [ int | float ] ] | list [ np.ndarray ] 
			A list of axes pointing into the directions of the Lattice.
		repetitions : int | list [ int ] | np.ndarray
			A list of repetitions for each axis with the same length as ``directions``
		name : str | None
			The name of the Lattice (used for displaying purposes).
		

		Returns
		-------
		blue.LatticeType
			A Lattice created with the MoveableThing as components
		"""
		
		return blue.Lattice(thing=self, 
				    directions=directions, 
				    repetitions=repetitions, 
				    name=name)

	# UTILITY METHODS

	@property
	def _location_range(self) -> tuple[np.ndarray, np.ndarray]|None:
		"""
		This function is used to determine the range in which the Thing and its children are located.
		
		Returns
		-------
		tuple[np.ndarray, np.ndarray]
			The returned values are two numpy arrays of shape (3,) containing the 
			minimum and maximum of the coordinate values of each dimension which 
			contain all points of the Thing including its children.
		"""
		if hasattr(self, 'size'):
			pos = blue.geometry.Vector.global_position(self)
			max_size = np.max(np.abs(self.size))
			min_pos = pos - max_size
			max_pos = pos + max_size
			return min_pos, max_pos


	def _update_rotation(self) -> None:
		"""
		This method updates the rotation data in the current mujoco simulation.
		"""
		if isinstance(self.parent, blue.WorldType):
			parent_R = np.eye(3)
		else:
			parent_R = self.parent.global_rotation_matrix
		R = blue.Rotation.E_rot(self._alpha, self._beta, self._gamma)
		global_R = parent_R @ R
		data_view = getattr(self.root._mj_data, self._MUJOCO_OBJ)
		data_view(self.name).xmat = global_R.reshape((-1,))


	@restrict
	@staticmethod
	def _get_pos(pos: list[int|float]|np.ndarray|blue.MoveableThingType) -> np.ndarray:
		"""
		This method is used to obtain the position from an input that is either a 
		numeric representation as in a list or a numpy array or a :class:`MoveableThing` 
		from which the position is returned.
		
		Parameters
		----------
		pos : list[int | float] | np.ndarray
			An object either 1D array like object or a :class:`MoveableThing`
		
		Returns
		-------
		np.ndarray
			The position used for further processing.
		"""
		if isinstance(pos, blue.MoveableThingType):
			pos = pos.global_pos
		return np.array(pos, dtype=np.float32)


	@property
	def rotation_matrix(self) -> np.ndarray:
		"""
		The matrix is constructed for local orientation (w.r.t. to parent orientation) — in contrast to 
		global orientation. For this use :meth:`Rotation.global_orientation <blueprints.utils.geometry.Rotation.global_orientation>`.

		Returns
		-------
		np.ndarray
			The rotation matrix for the Thins orientation.
		"""
		if self._launched:
			if isinstance(self.parent, blue.WorldType):
				parent_R = np.eye(3)
			else:
				parent_R = self.parent.global_rotation_matrix
			return parent_R.T @ self.global_rotation_matrix
		else:
			return blue.geometry.Rotation.E_rot(self.alpha, self.beta, self.gamma)
	

	@property
	def global_rotation_matrix(self) -> np.ndarray:
		"""
		The matrix is constructed for global orientation) — in contrast to 
		local orientation. For this use :meth:`Rotation.E_rot <blueprints.utils.geometry.Rotation.Rotation.E_rot>`.

		Returns
		-------
		np.ndarray
			The rotation matrix for the Thins orientation.
		"""
		if self._launched:
			#return self.root._mj_data.__getattibute__(self._MUJOCO_OBJ)(self.name).xmat.reshape((3, 3))
			data_view = getattr(self.root._mj_data, self._MUJOCO_OBJ)
			return data_view(self.name).xmat.reshape((3, 3))
		else:
			return blue.geometry.Rotation.global_rotation_matrix(self)

	# TRANSFORMATION METHODS

	@restrict
	def shift(self, 
		  pos:      list[int|float]|np.ndarray|blue.MoveableThingType|None = None, 
		  globally: bool = False, 
		  x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  **kwargs) -> blue.MoveableThingType:
		"""
		This method creates a copy that is shifted by ``pos``.

		>>> MoveableThing(pos=[5, 0, 0]).shift(x=-3, y=10).pos
		array([2., 10., 0.], dtype=float32)

		Parameters
		----------
		pos : list[int | float] | np.ndarray | blue.MoveableThingType | None
			The amount by which the Thing is shifted, optionally another :class:`MoveableThing` 
			can be passed in which case the position of this Thing is taken.
		globally : bool, optional
			Determining whether the shift is relative to the local orientation or the global orientation.
		x : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the X component of the position is to be changed.
		y : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Y component of the position is to be changed.
		z : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Z component of the position is to be changed.
		**kwargs
			Keyword arguments are passed to the :meth:`copy`.
		
		Returns
		-------
		blue.MoveableThingType
			A shifted copy is returned.
		"""
		if pos is not None and any(arg is not None for arg in [x, y, z]):
			raise ValueError(f'Only argument (pos) or arguments (x, y, z) can be set. Method call got ({', '.join(name for name, val in [('pos', pos), ('x', x), ('y', y), ('z', z)] if val is not None)}).')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		pos   = self._get_pos(pos)
		thing = self.copy(**kwargs)
		if globally:
			rotation_matrix = blue.geometry.Rotation.global_rotation_matrix(self)
			pos = rotation_matrix.T @ pos
		thing.pos = thing.pos + pos
		return thing


	@restrict
	def locate(self, 
		   pos:      list[int|float]|np.ndarray|blue.MoveableThingType|None = None, 
		   globally: bool = False, 
		   x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   **kwargs) -> blue.MoveableThingType:
		"""
		This method creates a copy that is relocated to ``pos``.
		
		>>> MoveableThing(pos=[5, 0, 0]).locate(x=-3, y=10).pos
		array([-3., 10., 0.], dtype=float32)

		Parameters
		----------
		pos : list[int | float] | np.ndarray | blue.MoveableThingType | None
			The position to which the Thing is shifted, optionally another :class:`MoveableThing` 
			can be passed in which case the position of this Thing is taken.
		globally : bool, optional
			Determining whether the relocation is relative to the local orientation or the global orientation.
		x : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the ``X`` component of the position is to be changed.
		y : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the ``Y`` component of the position is to be changed.
		z : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the ``Z`` component of the position is to be changed.
		**kwargs
			Keyword arguments are passed to :meth:`copy`.
		
		Returns
		-------
		blue.MoveableThingType
			A relocated copy is returned.
		"""
		if pos is not None and sum(arg is not None for arg in [x, y, z]) > 0:
			raise ValueError(f'Only argument (pos) or arguments (x, y, z) can be set. Method call got ({', '.join(name for name, val in [('pos', pos), ('x', x), ('y', y), ('z', z)] if val is not None)}).')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		pos   = self._get_pos(pos)
		thing = self.copy(**kwargs)
		if globally:
			rotation_matrix = blue.geometry.Rotation.global_rotation_matrix(self)
			pos = rotation_matrix.T @ pos
		thing.pos = np.array(pos)
		return thing


	@restrict
	def rotate(self, 
		   alpha:    int|float = 0, 
		   beta:     int|float = 0, 
		   gamma:    int|float = 0, 
		   quat:     list[int|float]|np.ndarray|None = None, 
		   globally: bool      = True, 
		   center:   list[int|float]|np.ndarray|blue.MoveableThingType|None = None, 
		   **kwargs) -> blue.MoveableThingType:
		"""
		This method creates a copy that is rotated by the (improper) euler angles.

		>>> MoveableThing(alpha=TAU/4, beta=0, gamma=TAU/8).rotate(beta=PI).euler
		array([-TAU/4,  PI,  TAU/8], dtype=float32)

		Parameters
		----------
		alpha : int | float, optional
			The (improer) euler angle of rotation around the ``X``-axis in radians.
		beta : int | float, optional
			The (improer) euler angle of rotation around the ``Y``-axis in radians.
		gamma : int | float, optional
			The (improer) euler angle of rotation around the ``Z``-axis in radians.
		globally : bool, optional
			Determining whether the rotation is relative to the local orientation or the global orientation.
		center : list[int | float] | np.ndarray | blue.MoveableThingType | None, optional
			The copied Thing will be rotated around center. If center is None, it will be rotated at its position instead.
		**kwargs
			Keyword arguments are passed to the :meth:`copy`.
		
		Returns
		-------
		blue.MoveableThingType
			A copy of the returned Thing is returned.
		"""
		thing = self.copy(**kwargs)
		if quat is not None:
			R, I, J, K = quat[0], quat[1], quat[2], quat[3]
			alpha, beta, gamma = blue.Rotation.quat_to_euler(R, I, J, K)
		if globally:
			R1 = blue.geometry.Rotation.global_rotation_matrix(self)
			R2 = blue.geometry.Rotation.E_rot(alpha, beta, gamma)
			R  = R2 @ R1
		else:
			R1 = thing.rotation_matrix
			R2 = blue.geometry.Rotation.E_rot(alpha, beta, gamma)
			R  = R1 @ R2
		alpha, beta, gamma = blue.geometry.Rotation.reference_frame_to_euler(R)
		thing.alpha = alpha
		thing.beta  = beta
		thing.gamma = gamma
		if center is not None:
			center = self._get_pos(center)
			if globally:
				thing.global_pos = R2 @ (thing.global_pos - center) + center
			else:
				thing.pos = R2 @ (thing.pos - center) + center
		return thing


	@restrict
	def align(self, 
		  alpha:    int|float = 0., 
		  beta:     int|float = 0., 
		  gamma:    int|float = 0., 
		  quat:     list[int|float]|np.ndarray|None = None, 
		   **kwargs) -> blue.MoveableThingType:
		"""
		This method creates a copy that is oriented according to the (improper) euler angles.
		
		>>> MoveableThing(alpha=TAU/4, beta=0, gamma=TAU/8).align(beta=PI).euler
		array([0,  PI,  0], dtype=float32)

		Parameters
		----------
		alpha : int | float, optional
			The (improer) euler angle of rotation around the ``X``-axis in radians.
		beta : int | float, optional
			The (improer) euler angle of rotation around the ``Y``-axis in radians.
		gamma : int | float, optional
			The (improer) euler angle of rotation around the ``Z``-axis in radians.
		**kwargs
			Keyword arguments are passed to the :meth:`copy`.
		
		Returns
		-------
		blue.MoveableThingType
			A copy of the Thing set to the given orientation is returned.
		"""
		thing = self.copy(**kwargs)
		if quat is not None:
			R, I, J, K = quat[0], quat[1], quat[2], quat[3]
			alpha, beta, gamma = blue.Rotation.quat_to_euler(R, I, J, K)
		thing.alpha = alpha if alpha is not None else 0
		thing.beta  = beta  if beta  is not None else 0
		thing.gamma = gamma if gamma is not None else 0
		return thing


	@restrict
	def scaled(self, scale: int|float, keep_pos: bool = False, **kwargs) -> blue.MoveableThingType:
		"""
		This method creates a copy that is rescaled in size and position.

		scale : int | float | None, optional
			The scaling factor by which the Things size and position is increased.
		keep_pos : bool
			If set, the position is kept, otherwise, the position is scaled as well.
		**kwargs
			Keyword arguments are passed to the :meth:`copy`.
		
		Returns
		-------
		blue.MoveableThingType
			A copy of the rescaled Thing.
		"""
		if isinstance(self, blue.NodeThingType):
			children  = {name: [thing.scaled(scale, keep_pos) for thing in list(child_dict['children'])] \
									 for name, child_dict in self._CHILDREN.items() \
									 if name not in kwargs and not issubclass(child_dict['type'], blue.CyclicalThingType)}
			cyclicals = {name: [thing.scaled(scale, keep_pos) for thing in list(child_dict['children'])] \
									 for name, child_dict in self._CHILDREN.items() \
									 if name not in kwargs and     issubclass(child_dict['type'], blue.CyclicalThingType)}
			blueprint_specs = self._blueprint_specs()
			blueprint_specs.update(children)
			blueprint_specs.update(kwargs)
			if 'name' not in kwargs:
				blueprint_specs['name'] = self._name
			thing = self.__class__(copy=False, **blueprint_specs)
			self._migrate_children(thing, cyclicals)
		else:
			thing     = self.copy(**kwargs)
		if not keep_pos:
			thing.pos = thing.pos * scale
		if hasattr(thing, 'size'):
			thing.size = thing.size * scale
		return thing

	# MUJOCO PROPERTIES

	@property
	def euler(self) -> np.ndarray:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Returns
		-------
		np.ndarray
			The orientation is represented as a vector containing the (improper) euler 
			angles array([alpha, beta, gamma]).
		"""
		if self._launched:
			R = self.rotation_matrix
			alpha, beta, gamma = blue.Rotation.reference_frame_to_euler(R)
			return np.array([alpha, 
					 beta,
					 gamma], dtype=np.float32)
		else:
			return np.array([self.alpha, 
					 self.beta,
					 self.gamma], dtype=np.float32)


	@euler.setter
	@restrict
	def euler(self, value: list[int|float]|np.ndarray) -> None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Parameters
		----------
		value : list[int | float] | np.ndarray
			The orientation is represented as a vector containing the (improper) euler 
			angles array([alpha, beta, gamma]).

		Raises
		------
		ValueError
			If the (improper) euler angles do not have the correct shape an error is raised.
		"""
		if isinstance(value, np.ndarray) and value.shape != (3,):
			raise ValueError(f'Orientation attribute euler must have a dimension of 3 got {value.shape} instead.')
		if isinstance(value, list) and len(value) != 3:
			raise ValueError(f'Orientation attribute euler must have a dimension of 3 got {len(value)} instead.')
		self._alpha, self._beta, self._gamma = map(float, value)
		if self._launched:
			self._update_rotation()


	@property
	def pos(self) -> np.ndarray:
		"""
		Individual components of :attr:`pos` can be found in :attr:`x`, :attr:`y` and :attr:`z`.

		Returns
		-------
		np.ndarray
			The position of the Thing is given in local coordinates.
		"""
		if self._launched:
			if isinstance(self.parent, blue.WorldType):
				parent_pos = np.zeros(3)
				parent_R   = np.eye(3)
			else:
				parent_pos = self.parent.global_pos
				parent_R   = self.parent.global_rotation_matrix
			delta_pos = self.global_pos - parent_pos
			return parent_R.T @ delta_pos
		else:
			return self._pos.copy()


	@pos.setter
	@restrict
	def pos(self, pos: np.ndarray|list[int|float]) -> None:
		"""
		Individual components of :attr:`pos` can be found in :attr:`x`, :attr:`y` and :attr:`z`.

		Parameters
		----------
		pos : np.ndarray | list[int | float]
			The position is always set in local coordinates.
		
		Raises
		------
		ValueError
			If the position does not have the correct shape an error is raised.
		"""
		if isinstance(pos, np.ndarray) and pos.shape != (3,):
			raise ValueError(f'Position attribute pos must have a dimension of 3 got {pos.shape} instead.')
		if isinstance(pos, list) and len(pos) != 3:
			raise ValueError(f'Position attribute pos must have a dimension of 3 got {len(pos)} instead.')
		pos = np.array(pos, dtype=np.float32)
		if self._launched:
			array = self.root._mj_model.__getattribute__(f'{self._MUJOCO_DATA}_pos')
			array[self._index] = pos
			if isinstance(self, blue.BodyType):
				delta_pos  = pos - self.pos
				global_pos = self.global_pos
				if isinstance(self.parent, blue.WorldType):
					R = np.eye(3)
				else:
					R = self.parent.global_rotation_matrix
				self.root._mj_data.xpos[self._index] = global_pos + R @ delta_pos
		self._pos = pos


	@property
	def global_pos(self):
		"""
		Individual components of :attr:`global_pos` can be found in :attr:`global_x`, :attr:`global_y` and :attr:`global_z`.

		Returns
		-------
		np.ndarray
			The position of the Thing is given in global coordinates.
		"""
		if self._launched:
			pos = np.zeros(3)
			for node in self.path[::-1]:
				if isinstance(node, blue.BodyType):
					global_R   = blue.Rotation.global_rotation_matrix(node)
					global_pos = self.root._mj_data.xpos[node._index]
					pos = global_pos + global_R @ pos
					break
				else:
					pos = node._pos + node.rotation_matrix @ pos
			return pos
		else:
			return blue.Vector.global_position(self)


	@global_pos.setter
	@restrict
	def global_pos(self, pos: np.ndarray|list[int|float]) -> None:
		"""
		Individual components of :attr:`pos` can be found in :attr:`x`, :attr:`y` and :attr:`z`.

		Parameters
		----------
		pos : np.ndarray | list[int | float]
			The position is always set in local coordinates.
		
		Raises
		------
		ValueError
			If the position does not have the correct shape an error is raised.
		"""
		if isinstance(pos, np.ndarray) and pos.shape != (3,):
			raise ValueError(f'Position attribute pos must have a dimension of 3 got {pos.shape} instead.')
		if isinstance(pos, list) and len(pos) != 3:
			raise ValueError(f'Position attribute pos must have a dimension of 3 got {len(pos)} instead.')
		pos = np.array(pos, dtype=np.float32)
		delta_pos = pos - self.global_pos
		if isinstance(self.parent, blue.WorldType) or self.parent is None:
			global_R = np.eye(3)
		else:
			global_R  = self.parent.global_rotation_matrix
		if self._launched:
			array = self.root._mj_model.__getattribute__(f'{self._MUJOCO_DATA}_pos')
			array[self._index] += global_R.T @ delta_pos
			if isinstance(self, blue.BodyType):
				self.root._mj_data.xpos[self._index] = pos
		self._pos += global_R.T @ delta_pos


	@property
	def alpha(self) -> float|None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Returns
		-------
		float | None
			The (improper) euler angle of rotation around the X-axis.
		"""
		if self._launched:
			return float(self.euler[0])
		else:
			return self._alpha


	@alpha.setter
	@restrict
	def alpha(self, alpha: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Parameters
		----------
		alpha : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The (improper) euler angle of rotation around the X-axis.
		"""
		self._alpha = float(alpha)
		if self._launched:
			self._update_rotation()


	@property
	def beta(self) -> float|None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Returns
		-------
		float | None
			The (improper) euler angle of rotation around the Y-axis.
		"""
		if self._launched:
			return float(self.euler[1])
		else:
			return self._beta


	@beta.setter
	@restrict
	def beta(self, beta: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Parameters
		----------
		beta : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The (improper) euler angle of rotation around the Y-axis.
		"""
		self._beta = float(beta)
		if self._launched:
			self._update_rotation()


	@property
	def gamma(self) -> float|None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Returns
		-------
		float | None
			The (improper) euler angle of rotation around the Z-axis.
		"""
		if self._launched:
			return float(self.euler[2])
		else:
			return self._gamma


	@gamma.setter
	@restrict
	def gamma(self, gamma: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		Note that euler angles are not additive for rotations. Use :meth:`rotate` for this.
		
		Parameters
		----------
		gamma : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The (improper) euler angle of rotation around the Z-axis.
		"""
		self._gamma = float(gamma)
		if self._launched:
			self._update_rotation()


	@property
	def quat(self):
		"""
		Alternatively `quaternions <https://en.wikipedia.org/wiki/Quaternion>`__ can be used for orientations. 
		Where every euler angles are provided, a ``quat`` can be passed instead.

		Returns
		-------
		np.ndarray
		"""
		quat = blue.Rotation.euler_to_quat(self.alpha, self.beta, self.gamma)
		return np.array(quat, dtype=np.float32)


	@quat.setter
	@blue.restrict
	def quat(self, quat: list[int|float]|np.ndarray):
		R, I, J, K = quat[0], quat[1], quat[2], quat[3]
		alpha, beta, gamma = blue.Rotation.quat_to_euler(R, I, J, K)
		self.alpha = alpha
		self.beta  = beta
		self.gamma = gamma
		
	


	@property
	def x(self) -> float|None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Returns
		-------
		float | None
			The ``X``-coordinate of the Things local position.
		"""
		return float(self.pos[0])


	@x.setter
	@restrict
	def x(self, x: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Parameters
		----------
		x : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The X-coordinate of the Things local position.
		"""
		pos      = self.pos
		pos[0]   = float(x)
		self.pos = pos


	@property
	def y(self) -> float|None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Returns
		-------
		float | None
			The ``Y``-coordinate of the Things local position.
		"""
		return float(self.pos[1])


	@y.setter
	@restrict
	def y(self, y: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Parameters
		----------
		y : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The Y-coordinate of the Things local position.
		"""
		pos      = self.pos
		pos[1]   = float(y)
		self.pos = pos


	@property
	def z(self) -> float|None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Returns
		-------
		float | None
			The ``Z``-coordinate of the Things local position.
		"""
		return float(self.pos[2])


	@z.setter
	@restrict
	def z(self, z: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		If the whole position should be updated use :attr:`pos` instead.
		
		Parameters
		----------
		z : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The Z-coordinate of the Things local position.
		"""
		pos      = self.pos
		pos[2]   = float(z)
		self.pos = pos


	@property
	def vel(self) -> np.ndarray:
		"""
		The global velocity for all axis
		
		Returns
		-------
		np.ndarray
		"""
		if self._launched:
			node = self
			while not isinstance(node, blue.BodyType):
				node = node.parent
				if node is None:
					return np.zeros(3)
			vel = self.root._mj_data.body(node._index).cvel[3:]
			return np.array(vel)
		else:
			return np.zeros(3)


	@property
	def x_vel(self):
		"""
		The global velocity of the ``X`` axis
		
		Returns
		-------
		np.ndarray
		"""
		return self.vel[0]


	@property
	def y_vel(self):
		"""
		The global velocity of the ``Y`` axis
		
		Returns
		-------
		np.ndarray
		"""
		return self.vel[1]


	@property
	def z_vel(self):
		"""
		The global velocity of the ``Z`` axis
		
		Returns
		-------
		np.ndarray
		"""
		return self.vel[2]
	


	# NO EFFECT ON SIMULATION
	#@vel.setter
	#@restrict
	#def vel(self, vel: np.ndarray|list[int|float]) -> None:
	#	"""
	#	Parameters
	#	-------
	#	vel : np.ndarray | list[int | float]
	#		The global velocity for all axis
	#	"""
	#	if isinstance(vel, np.ndarray) and vel.shape != (3,):
	#		raise ValueError(f'Argument vel must be of shape (3,) but received was {vel.shape}!')
	#	elif len(vel) != 3:
	#		raise ValueError(f'Argument vel must be of shape (3,) but received was ({len(vel)},)!')
	#	if self._launched:
	#		node = self
	#		while not isinstance(node, blue.BodyType):
	#			node = node.parent
	#			if node is None:
	#				return
	#		self.root._mj_data.body(node._index).cvel[3:] = vel


	@property
	def angular_vel(self) -> np.ndarray:
		"""
		The global angular velocity for all axis
		
		Returns
		-------
		np.ndarray
		"""
		if self._launched:
			node = self
			while not isinstance(node, blue.BodyType):
				node = node.parent
				if node is None:
					return np.zeros(3)
			vel = self.root._mj_data.body(node._index).cvel[:3]
			return np.array(vel)
		else:
			return np.zeros(3)


	# NO EFFECT ON SIMULATION
	#@angular_vel.setter
	#@restrict
	#def angular_vel(self, angular_vel: np.ndarray|list[int|float]) -> None:
	#	"""
	#	Parameters
	#	-------
	#	angular_vel : np.ndarray | list[int | float]
	#		The global angular velocity for all axis
	#	"""
	#	if isinstance(angular_vel, np.ndarray) and angular_vel.shape != (3,):
	#		raise ValueError(f'Argument angular_vel must be of shape (3,) but received was {angular_vel.shape}!')
	#	elif len(angular_vel) != 3:
	#		raise ValueError(f'Argument angular_vel must be of shape (3,) but received was ({len(angular_vel)},)!')
	#	if self._launched:
	#		node = self
	#		while not isinstance(node, blue.BodyType):
	#			node = node.parent
	#			if node is None:
	#				return
	#		self.root._mj_data.body(node._index).cvel[:3] = angular_vel



"""
Things are base classes used to implement the rudimentary features of mujoco objects. All 
objects that are used to build a kinematic tree in blueprints are Things. A variety of 
Things can be created and combined in a kinematic structure to design your simulation. 
Most of the time you will create objects that are used multiple times in your simulation.
Therefore, once a Thing is created it serves as a blueprint. Attaching it to another Thing 
will create a copy, such that the blueprint is unchanged and can be altered and attached at 
different nodes in the kinematic tree without changing the the previously attach version 
of itself.

:class:`base.BaseThing` is the class from which all objects in the kinematic tree ultimately derive. 
It handles the basic methods used to translate from blueprints in python to mujoco xml.

:class:`node.NodeThing` is used to implement methods that are used to create structure for the 
kinematic tree. Every Thing that can have children attached to it is a :class:`node.NodeThing`.

:class:`moveable.MoveableThing` implements position and rotation attributes as well as methods to 
create rotated and shifted copies of itself.

:class:`colored.ColoredThing` contains attributes for color and opacity.

:class:`unique.UniqueThing` are objects that should be copied seldomly. Those are Things to which 
multiple other Things might refer and that should not be duplicated arbitrarily. Some Things 
like :class:`blueprints.geoms.Mesh` will appear to the user as one object, but are actually 
separated into multiple objects in the background to store large datasets separately. This 
enables memory efficiency for copies of the mesh that do not alter the dataset contained.

:class:`cyclical.CyclicalThing` are used for all Things that might refer to multiple other Things in 
such that it creates a graph cycle.

"""

from .base     import *
from .node     import *
from .moveable  import *
from .colored  import *
from .unique   import *
from .cyclical import *
